#!/usr/bin/env python
from __future__ import print_function

MODDOC = """
=== Epics Scanning Functions for Larch ===


This does not used the Epics SScan Record, and the scan is intended to run
as a python application, but many concepts from the Epics SScan Record are
borrowed.  Where appropriate, the difference will be noted here.

A Step Scan consists of the following objects:
   a list of Positioners
   a list of Triggers
   a list of Counters

Each Positioner will have a list (or numpy array) of position values
corresponding to the steps in the scan.  As there is a fixed number of
steps in the scan, the position list for each positioners must have the
same length -- the number of points in the scan.  Note that, unlike the
SScan Record, the list of points (not start, stop, step, npts) must be
given.  Also note that the number of positioners or number of points is not
limited.

A Trigger is simply an Epics PV that will start a particular detector,
usually by having 1 written to its field.  It is assumed that when the
Epics ca.put() to the trigger completes, the Counters associated with the
triggered detector will be ready to read.

A Counter is simple a PV whose value should be recorded at every step in
the scan.  Any PV can be a Counter, including waveform records.  For many
detector types, it is possible to build a specialized class that creates
many counters.

Because Triggers and Counters are closely associated with detectors, a
Detector is also defined, which simply contains a single Trigger and a list
of Counters, and will cover most real use cases.

In addition to the core components (Positioners, Triggers, Counters, Detectors),
a Step Scan contains the following objects:

   breakpoints   a list of scan indices at which to pause and write data
                 collected so far to disk.
   extra_pvs     a list of (description, PV) tuples that are recorded at
                 the beginning of scan, and at each breakpoint, to be
                 recorded to disk file as metadata.
   pre_scan()    method to run prior to scan.
   post_scan()   method to run after scan.
   at_break()    method to run at each breakpoint.

Note that Postioners and Detectors may add their own pieces into extra_pvs,
pre_scan(), post_scan(), and at_break().

With these concepts, a Step Scan ends up being a fairly simple loop, going
roughly (that is, skipping error checking) as:

   pos = <DEFINE POSITIONER LIST>
   det = <DEFINE DETECTOR LIST>
   run_pre_scan(pos, det)
   [p.move_to_start() for p in pos]
   record_extra_pvs(pos, det)
   for i in range(len(pos[0].array)):
       [p.move_to_pos(i) for p in pos]
       while not all([p.done for p in pos]):
           time.sleep(0.001)
       [trig.start() for trig in det.triggers]
       while not all([trig.done for trig in det.triggers]):
           time.sleep(0.001)
       [det.read() for det in det.counters]

       if i in breakpoints:
           write_data(pos, det)
           record_exrta_pvs(pos, det)
           run_at_break(pos, det)
   write_data(pos, det)
   run_post_scan(pos, det)

Note that multi-dimensional mesh scans over a rectangular grid is not
explicitly supported, but these can be easily emulated with the more
flexible mechanism of unlimited list of positions and breakpoints.
Non-mesh scans are also possible.

A step scan can have an Epics SScan Record or StepScan database associated
with it.  It will use these for PVs to post data at each point of the scan.
"""
import os, shutil
import time
from threading import Thread
import json
import numpy as np
import random

from datetime import timedelta

from larch import Group
from larch.utils import debugtime, fix_varname

try:
    from epics import PV, caget, caput, get_pv, poll
    HAS_EPICS = True
except ImportError:
    HAS_EPICS = False

try:
    import epicsscan
    from epicsscan import (Counter, Trigger, AreaDetector, get_detector,
                           ASCIIScanFile, Positioner)
    from epicsscan.scandb import ScanDB, InstrumentDB, ScanDBException, ScanDBAbort
    from epics.devices.struck import Struck

    HAS_EPICSSCAN = True
except ImportError:
    HAS_EPICSSCAN = False


SCANDB_NAME = '_scan._scandb'
INSTDB_NAME = '_scan._instdb'
_scandb = None
_instdb = None

def scan_from_db(scanname, filename='scan.001',  _larch=None):
    """
    get scan definition from ScanDB by name
    """
    global _scandb
    if _scandb is None:
        print('scan_from_db: need to connect to scandb!')
        return
    try:
        scan = _scandb.make_scan(scanname, larch=_larch)
        scan.filename = filename
    except ScanDBException:
        raise ScanDBException("no scan definition '%s' found" % scanname)
    return scan

def do_scan(scanname, filename='scan.001', nscans=1, comments='', _larch=None):
    """do_scan(scanname, filename='scan.001', nscans=1, comments='')

    execute a step scan as defined in Scan database

    Parameters
    ----------
    scanname:     string, name of scan
    filename:     string, name of output data file
    comments:     string, user comments for file
    nscans:       integer (default 1) number of repeats to make.

    Examples
    --------
      do_scan('cu_xafs', 'cu_sample1.001', nscans=3)

    Notes
    ------
      1. The filename will be incremented so that each scan uses a new filename.
    """
    global _scandb
    if _scandb is None:
        print('do_scan: need to connect to scandb!')
        return
    if nscans is not None:
        _scandb.set_info('nscans', nscans)

    scan = scan_from_db(scanname, filename=filename,  _larch=_larch)
    scan.comments = comments
    if scan.scantype == 'slew':
        return scan.run(filename=filename, comments=comments)
    else:
        scans_completed = 0
        nscans = int(_scandb.get_info('nscans'))
        abort  = _scandb.get_info('request_abort', as_bool=True)
        while (scans_completed  < nscans) and not abort:
            scan.run()
            scans_completed += 1
            nscans = int(_scandb.get_info('nscans'))
            abort  = _scandb.get_info('request_abort', as_bool=True)
        return scan

def get_dbinfo(key, default=None, as_int=False, as_bool=False,
               full_row=False, _larch=None, **kws):
    """get a value for a keyword in the scan info table,
    where most status information is kept.

    Arguments
    ---------
     key        name of data to look up
     default    (default None) value to return if key is not found
     as_int     (default False) convert to integer
     as_bool    (default False) convert to bool
     full_row   (default False) return full row, not just value

    Notes
    -----
     1.  if this key doesn't exist, it will be added with the default
         value and the default value will be returned.
     2.  the full row will include notes, create_time, modify_time

    """
    global _scandb
    if _scandb is None:
        print('get_dbinfo: need to connect to scandb!')
        return
    return _scandb.get_info(key, default=default, full_row=full_row,
                            as_int=as_int, as_bool=as_bool, **kws)

def set_dbinfo(key, value, notes=None, _larch=None, **kws):
    """
    set a value for a keyword in the scan info table.
    """
    global _scandb
    if _scandb is None:
        print('set_dbinfo: need to connect to scandb!')
        return
    return _scandb.set_info(key, value, notes=notes)

def connect_scandb(scandb=None, dbname=None, _larch=None, **kwargs):
    from epicsscan.scandb import ScanDB, InstrumentDB
    global _scandb, _instdb
    if _scandb is not None:
        return _scandb
    if scandb is None:
        scandb = ScanDB(dbname=dbname, **kwargs)

    if scandb is not None:
        _scandb = scandb

    if _larch is not None:
        _larch.symtable.set_symbol(SCANDB_NAME, _scandb)

    if _instdb is None:
        _instdb = InstrumentDB(_scandb)

        if _larch is not None:
            _larch.symtable.set_symbol(INSTDB_NAME, _instdb)
    return _scandb



#!/usr/bin/env python
"""
"""
import os
import time
import shutil
import numpy as np
from random import randrange
from functools import partial
from datetime import timedelta

import wx
import wx.lib.scrolledpanel as scrolled
import wx.lib.mixins.inspection

HAS_EPICS = False
try:
    import epics
    from epics.wx import DelayedEpicsCallback, EpicsFunction
    HAS_EPICS = True
except ImportError:
    pass

import larch
from ..larchlib import read_workdir, save_workdir

from ..io  import (gsescan_deadtime_correct, gsexdi_deadtime_correct,
                   is_GSEXDI, AthenaProject, new_filename, increment_filename)

from wxutils import (SimpleText, FloatCtrl, pack, Button, Popup,
                     Choice,  Check, MenuItem, GUIColors,
                     CEN, LEFT, FRAMESTYLE, Font)

CEN |=  wx.ALL
FILE_WILDCARDS = "Scan Data Files(*.0*,*.1*,*.dat,*.xdi)|*.0*;*.1*;*.dat;*.xdi|All files (*)|*"


def okcancel(panel, onOK=None, onCancel=None):
    btnsizer = wx.StdDialogButtonSizer()
    _ok = wx.Button(panel, wx.ID_OK)
    _no = wx.Button(panel, wx.ID_CANCEL)
    panel.Bind(wx.EVT_BUTTON, onOK,     _ok)
    panel.Bind(wx.EVT_BUTTON, onCancel, _no)
    _ok.SetDefault()
    btnsizer.AddButton(_ok)
    btnsizer.AddButton(_no)
    btnsizer.Realize()
    return btnsizer

class DTCorrectFrame(wx.Frame):
    _about = """GSECARS Deadtime Corrections
  Matt Newville <newville @ cars.uchicago.edu>
  """
    def __init__(self, _larch=None, **kws):

        wx.Frame.__init__(self, None, -1, style=FRAMESTYLE)
        self.file_groups = {}
        self.file_paths  = []
        title = "DeadTime Correction "
        self.larch = _larch
        self.subframes = {}

        self.SetSize((500, 275))
        self.SetFont(Font(10))

        self.config = {'chdir_on_fileopen': True}
        self.SetTitle(title)
        self.createMainPanel()
        self.createMenus()
        self.statusbar = self.CreateStatusBar(2, 0)
        self.statusbar.SetStatusWidths([-3, -1])
        statusbar_fields = ["Initializing....", " "]
        for i in range(len(statusbar_fields)):
            self.statusbar.SetStatusText(statusbar_fields[i], i)

    def onBrowse(self, event=None):
        dlg = wx.FileDialog(parent=self,
                        message='Select Files',
                        defaultDir=os.getcwd(),
                        wildcard =FILE_WILDCARDS,
                        style=wx.FD_OPEN|wx.FD_MULTIPLE|wx.FD_CHANGE_DIR)

        if dlg.ShowModal() == wx.ID_OK:
            path = dlg.GetPath()
            mdir, p = os.path.split(path)
            os.chdir(mdir)
            roiname = self.wid_roi.GetValue().strip()
            if len(roiname) < 1:
                Popup(self,
                    'Must give ROI name!', 'No ROI name')
                return
            dirname = self.wid_dir.GetValue().strip()
            if len(dirname) > 1 and not os.path.exists(dirname):
                try:
                    os.mkdir(dirname)
                except:
                    Popup(self,
                        'Could not create directory %s' % dirname,
                        "could not create directory")
                    return
            badchans = self.wid_bad.GetValue().strip()
            bad_channels = []
            if len(badchans)  > 0:
                bad_channels = [int(i.strip()) for i in badchans.split(',')]

            groups = []
            for fname in dlg.GetFilenames():
                corr_fcn = gsescan_deadtime_correct
                if is_GSEXDI(fname):
                    corr_fcn = gsexdi_deadtime_correct
                self.write_message("Correcting %s" % (fname))
                out = corr_fcn(fname, roiname, subdir=dirname,
                               bad=bad_channels, _larch=self.larch)
                if out is not None:
                    out.mu = out.mufluor
                    out.filename = fname
                    groups.append((out, fname))

            athena_name = os.path.join(dirname, self.wid_ath.GetValue().strip())
            if self.wid_autoname.IsChecked():
                athena_name = new_filename(athena_name)

            _, aname = os.path.split(athena_name)
            self.wid_ath.SetValue(increment_filename(aname))

            aprj = AthenaProject(filename=athena_name, _larch=self.larch)
            for grp, label in groups:
                aprj.add_group(grp, signal='mu')
            aprj.save(use_gzip=True)
            self.write_message("Corrected %i files, wrote %s" % (len(groups), aname))

    def createMainPanel(self):
        panel = wx.Panel(self)
        sizer = wx.GridBagSizer(5, 4)

        lab_roi = SimpleText(panel, ' Element / ROI Name:')
        lab_dir = SimpleText(panel, ' Output Folder:')
        lab_ath = SimpleText(panel, ' Athena Project File:')
        lab_bad = SimpleText(panel, ' Bad Channels:')
        lab_sel = SimpleText(panel, ' Select Files:')

        self.wid_roi = wx.TextCtrl(panel, -1, '', size=(200, -1))
        self.wid_dir = wx.TextCtrl(panel, -1, 'DT_Corrected', size=(200, -1))
        self.wid_ath = wx.TextCtrl(panel, -1, 'Athena_001.prj', size=(200, -1))
        self.wid_bad = wx.TextCtrl(panel, -1, ' ', size=(200, -1))
        self.wid_autoname = Check(panel, default=True,
                                  size=(150, -1), label='auto-increment?')

        self.sel_wid = Button(panel, 'Browse', size=(100, -1),
                                action=self.onBrowse)

        ir = 0
        sizer.Add(lab_roi,       (ir, 0), (1, 1), LEFT, 2)
        sizer.Add(self.wid_roi, (ir, 1), (1, 1), LEFT, 2)
        ir += 1
        sizer.Add(lab_dir,       (ir, 0), (1, 1), LEFT, 2)
        sizer.Add(self.wid_dir,  (ir, 1), (1, 1), LEFT, 2)
        ir += 1
        sizer.Add(lab_ath,       (ir, 0), (1, 1), LEFT, 2)
        sizer.Add(self.wid_ath,  (ir, 1), (1, 1), LEFT, 2)
        sizer.Add(self.wid_autoname,  (ir, 2), (1, 1), LEFT, 2)

        ir += 1
        sizer.Add(lab_bad,            (ir, 0), (1, 1), LEFT, 2)
        sizer.Add(self.wid_bad,  (ir, 1), (1, 1), LEFT, 2)
        ir += 1
        sizer.Add(lab_sel,       (ir, 0), (1, 1), LEFT, 2)
        sizer.Add(self.sel_wid,  (ir, 1), (1, 1), LEFT, 2)

        pack(panel, sizer)
        wx.CallAfter(self.init_larch)
        return

    def init_larch(self):
        t0 = time.time()
        if self.larch is None:
            self.larch = larch.Interpreter()
        self.larch.symtable.set_symbol('_sys.wx.wxapp', wx.GetApp())
        self.larch.symtable.set_symbol('_sys.wx.parent', self)
        self.SetStatusText('ready')

    def write_message(self, s, panel=0):
        """write a message to the Status Bar"""
        self.SetStatusText(s, panel)

    def createMenus(self):
        # ppnl = self.plotpanel
        self.menubar = wx.MenuBar()
        fmenu = wx.Menu()
        MenuItem(self, fmenu, "&Quit\tCtrl+Q", "Quit program", self.onClose)
        self.menubar.Append(fmenu, "&File")
        self.SetMenuBar(self.menubar)

    def onClose(self,evt):
        self.Destroy()


class DTViewer(wx.App, wx.lib.mixins.inspection.InspectionMixin):
    def __init__(self, _larch=None, **kws):
        self._larch = _larch
        wx.App.__init__(self, **kws)

    def run(self):
        self.MainLoop()

    def createApp(self):
        frame = DTCorrectFrame(_larch=self._larch)
        frame.Show()
        self.SetTopWindow(frame)

    def OnInit(self):
        self.createApp()
        return True

def dtcorrect(wxparent=None, _larch=None,  **kws):
    s = DTCorrectFrame(_larch=_larch, **kws)
    s.Show()
    s.Raise()



#!/usr/bin/env python
"""
Linear Combination panel
"""
import os
import time
import wx
import wx.lib.scrolledpanel as scrolled
import wx.dataview as dv

import numpy as np

from functools import partial
from collections import OrderedDict

from lmfit.printfuncs import gformat

from larch import Group
from larch.math import index_of
from larch.wxlib import (BitmapButton, FloatCtrl, get_icon, SimpleText,
                         pack, Button, HLine, Choice, Check, CEN, RIGHT,
                         LEFT, Font)

from .taskpanel import TaskPanel, autoset_fs_increment, DataTableGrid
from larch.math.lincombo_fitting import get_arrays

np.seterr(all='ignore')

# plot options:
norm   = 'Normalized \u03bC(E)'
dmude  = 'd\u03bC(E)/dE'
chik   = '\u03c7(k)'
noplot = '<no plot>'
noname = '<none>'

DVSTYLE = dv.DV_SINGLE|dv.DV_VERT_RULES|dv.DV_ROW_LINES

FitSpace_Choices = [norm, dmude, chik]
Plot_Choices = ['PCA Components', 'Component Weights', 'Data + Fit',
                'Data + Fit + Components']

defaults = dict(xmin=-5.e5, xmax=5.e5, fitspace=norm, weight_min=0.002,
                weight_auto=True, max_components=50)

# max number of *reported* PCA weights after fit
MAX_ROWS = 50

class PCAPanel(TaskPanel):
    """PCA Panel"""

    def __init__(self, parent, controller, **kws):
        TaskPanel.__init__(self, parent, controller,
                           configname='pca_config',
                           config=defaults,
                           title='Principal Component Analysis',
                           **kws)
        self.result = None

    def process(self, dgroup, **kws):
        """ handle PCA processing"""
        if self.skip_process:
            return
        form = self.read_form()

    def build_display(self):
        panel = self.panel
        wids = self.wids
        self.skip_process = True

        wids['fitspace'] = Choice(panel, choices=FitSpace_Choices, size=(175, -1))
        wids['fitspace'].SetStringSelection(norm)
        wids['plotchoice'] = Choice(panel, choices=Plot_Choices,
                                  size=(175, -1), action=self.onPlot)
        wids['plotchoice'].SetSelection(0)

        add_text = self.add_text

        opts = dict(digits=2, increment=1.0)

        w_xmin = self.add_floatspin('xmin', value=defaults['xmin'], **opts)
        w_xmax = self.add_floatspin('xmax', value=defaults['xmax'], **opts)

        w_wmin = self.add_floatspin('weight_min', digits=4,
                                    value=defaults['weight_min'],
                                    increment=0.001, with_pin=False,
                                    min_val=0,  max_val=0.5,
                                    action=self.onSet_WeightMin)

        autoset_fs_increment(self.wids['weight_min'], defaults['weight_min'])


        self.wids['weight_auto'] = Check(panel, default=True, label='auto?')


        w_mcomps = self.add_floatspin('max_components', digits=0,
                                    value=defaults['max_components'],
                                    increment=1, with_pin=False, min_val=0)

        b_build_model = Button(panel, 'Build Model With Selected Groups',
                                     size=(275, -1),  action=self.onBuildPCAModel)

        wids['fit_group'] = Button(panel, 'Test Current Group with Model', size=(225, -1),
                                   action=self.onFitGroup)

        wids['save_model'] = Button(panel, 'Save PCA Model', size=(150, -1),
                                    action=self.onSavePCAModel)
        wids['load_model'] = Button(panel, 'Load PCA Model', size=(150, -1),
                                    action=self.onLoadPCAModel)

        wids['fit_group'].Disable()
        wids['load_model'].Disable()
        wids['save_model'].Disable()

        collabels = [' Variance ', ' IND value ', ' IND / IND_Best',
                     ' Factor Weight' ]

        colsizes = [100, 100, 100, 200]
        coltypes = ['float:12,6', 'float:12,6', 'float:12,5', 'string']
        coldefs  = [0.0, 0.0, 1.0, '0.0']

        wids['table'] = DataTableGrid(panel, nrows=MAX_ROWS,
                                      collabels=collabels,
                                      datatypes=coltypes,
                                      defaults=coldefs,
                                      colsizes=colsizes, rowlabelsize=60)

        wids['table'].SetMinSize((625, 175))
        wids['table'].EnableEditing(False)
#
#         sview = self.wids['stats'] = dv.DataViewListCtrl(panel, style=DVSTYLE)
#         sview.AppendTextColumn(' Component',    width=75)
#         sview.AppendTextColumn(' Weight',       width=100)
#         sview.AppendTextColumn(' Significant?', width=75)
#
#         for col in range(sview.ColumnCount):
#             this = sview.Columns[col]
#             align = wx.ALIGN_LEFT if col == 0 else wx.ALIGN_RIGHT
#             this.Sortable = False
#             this.Alignment = this.Renderer.Alignment = align
#         sview.SetMinSize((275, 250))

        wids['status'] = SimpleText(panel, ' ')
        rfont = self.GetFont()
        rfont.SetPointSize(rfont.GetPointSize()+1)

        wids['fit_chi2'] = SimpleText(panel,   ' N/A ', size=(150,-1))
        wids['fit_dscale'] = SimpleText(panel, ' N/A ', size=(150,-1))

        panel.Add(SimpleText(panel, ' Principal Component Analysis',
                             **self.titleopts), dcol=4)

        add_text('Array to Use: ', newrow=True)
        panel.Add(wids['fitspace'], dcol=2)
        panel.Add(wids['load_model'], dcol=2)

        add_text('Plot : ', newrow=True)
        panel.Add(wids['plotchoice'], dcol=2)

        panel.Add(wids['save_model'], dcol=2)

        add_text('Fit Energy Range: ')
        panel.Add(w_xmin)
        add_text(' : ', newrow=False)
        panel.Add(w_xmax)
        #panel.Add(wids['show_fitrange'])

        add_text('Min Weight: ')
        panel.Add(w_wmin)
        panel.Add(wids['weight_auto'])
        add_text('Max Components:', dcol=1, newrow=True)
        panel.Add(w_mcomps)

        panel.Add(Button(panel, 'Copy To Selected Groups', size=(200, -1),
                         action=partial(self.onCopyParam, 'pca')),
                  dcol=2)

        add_text('Status: ')
        panel.Add(wids['status'], dcol=3)

        panel.Add(HLine(panel, size=(550, 2)), dcol=5, newrow=True)

        panel.Add(b_build_model, dcol=5, newrow=True)

        panel.Add(wids['table'], dcol=5, newrow=True)
        panel.Add(wids['fit_group'], dcol=3, newrow=True)

        add_text('chi-square: ')
        panel.Add(wids['fit_chi2'], dcol=3)
        add_text('scale factor: ')
        panel.Add(wids['fit_dscale'], dcol=3)

#
#         ## add weights report: slightly tricky layout
#         ## working with GridBagSizer under gridpanel...
#         icol = panel.icol - 2
#         irow = panel.irow
#         pstyle, ppad = panel.itemstyle, panel.pad
#
#         panel.sizer.Add(SimpleText(panel, 'data scalefactor = '),
#                         (irow+1, icol), (1, 1), pstyle, ppad)
#         panel.sizer.Add(wids['fit_dscale'],
#                         (irow+1, icol+1), (1, 1), pstyle, ppad)
#         irow +=1
#         for i in range(1, NWTS+1):
#             wids['fit_wt%d' % i] = SimpleText(panel, '--', font=rfont)
#             panel.sizer.Add(SimpleText(panel, 'weight_%d ='%i, font=rfont),
#                             (irow+i, icol), (1, 1), pstyle, ppad)
#             panel.sizer.Add(wids['fit_wt%d'%i],
#                             (irow+i, icol+1), (1, 1), pstyle, ppad)
#
        panel.pack()
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add((10, 10), 0, LEFT, 3)
        sizer.Add(panel, 1, LEFT, 3)
        pack(self, sizer)
        self.skip_process = False

    def onSet_WeightMin(self, evt=None, value=None):
        "handle setting edge step"
        wmin = self.wids['weight_min'].GetValue()
        self.wids['weight_auto'].SetValue(0)
        self.update_config({'weight_min': wmin})
        autoset_fs_increment(self.wids['weight_min'], wmin)

    def fill_form(self, dgroup):
        opts = self.get_config(dgroup)
        self.dgroup = dgroup
        if isinstance(dgroup, Group):
            d_emin = min(dgroup.energy)
            d_emax = max(dgroup.energy)
            if opts['xmin'] < d_emin:
                opts['xmin'] = -40 + int(dgroup.e0/10.0)*10
            if opts['xmax'] > d_emax:
                opts['xmax'] =  110 + int(dgroup.e0/10.0)*10

        self.skip_process = True
        wids = self.wids
        for attr in ('xmin', 'xmax', 'weight_min'):
            val = opts.get(attr, None)
            if val is not None:
                wids[attr].SetValue(val)

        for attr in ('fitspace',):
            if attr in opts:
                wids[attr].SetStringSelection(opts[attr])

        self.skip_process = False

    def onCopyParam(self, name=None, evt=None):
        conf = self.get_config()
        conf.update(self.read_form())
        attrs =  ('xmin', 'xmax', 'weight_min',
                  'max_components', 'fitspace', 'plotchoice')

        out = {a: conf[a] for a in attrs}
        for checked in self.controller.filelist.GetCheckedStrings():
            groupname = self.controller.file_groups[str(checked)]
            dgroup = self.controller.get_group(groupname)
            self.update_config(out, dgroup=dgroup)


    def plot_pca_weights(self, win=2):
        if self.result is None or self.skip_plotting:
            return
        form = self.read_form()
        cmd = "plot_pca_weights(pca_result, ncomps=%d, min_weight=%.3f, win=%d)"
        max_comps = form['max_components']
        min_weight = form['weight_min']
        self.larch_eval(cmd % (max_comps, min_weight, win))

    def plot_pca_components(self, win=1):
        if self.result is None or self.skip_plotting:
            return
        form = self.read_form()
        cmd = "plot_pca_components(pca_result, ncomps=%d, min_weight=%.3f, win=%d)"
        max_comps = form['max_components']
        min_weight = form['weight_min']
        self.larch_eval(cmd % (max_comps, min_weight, win))

    def plot_pca_fit(self, win=1):
        if self.result is None or self.skip_plotting:
            return
        form = self.read_form()
        with_comps = repr('components' in form['plotchoice'].lower())
        dgroup = self.controller.get_group()
        if hasattr(dgroup, 'pca_result'):
            a = (dgroup.groupname, win, with_comps)
            self.larch_eval("plot_pca_fit(%s, win=%d, with_components=%s)" % a)

    def onPlot(self, event=None):
        form = self.read_form()
        if self.skip_plotting:
            return
        pchoice = form['plotchoice'].lower()
        if pchoice.startswith('pca'):
            self.plot_pca_components()
        elif pchoice.startswith('component w'):
            self.plot_pca_weights()
        else:
            self.plot_pca_fit()

    def onFitGroup(self, event=None):
        form = self.read_form()
        if self.result is None:
            print("need result first!")
        ncomps = int(form['max_components'])
        gname = form['groupname']
        cmd = "pca_fit(%s, pca_result, ncomps=%d)" % (gname, ncomps)
        self.larch_eval(cmd)
        dgroup = self.controller.get_group()
        pca_chisquare = dgroup.pca_result.chi_square
        self.wids['fit_chi2'].SetLabel(gformat(dgroup.pca_result.chi_square))
        self.wids['fit_dscale'].SetLabel(gformat(dgroup.pca_result.data_scale))

        grid_data = self.wids['table'].table.data
        for g in grid_data: g[3] = '-'

        for i, wt in enumerate(dgroup.pca_result.weights):
            grid_data[i][3] = gformat(wt)

        self.wids['table'].table.data = grid_data
        self.wids['table'].table.View.Refresh()
        self.plot_pca_fit()

    def onBuildPCAModel(self, event=None):
        self.wids['status'].SetLabel(" training model...")
        form = self.read_form()
        selected_groups = self.controller.filelist.GetCheckedStrings()
        groups = [self.controller.file_groups[cn] for cn in selected_groups]
        for gname in groups:
            grp = self.controller.get_group(gname)
            if not hasattr(grp, 'norm'):
                self.parent.nb_panels[0].process(grp)

        groups = ', '.join(groups)
        opts = dict(groups=groups, arr='norm', xmin=form['xmin'], xmax=form['xmax'])
        cmd = "pca_result = pca_train([{groups}], arrayname='{arr}', xmin={xmin:.2f}, xmax={xmax:.2f})"

        self.larch_eval(cmd.format(**opts))
        r = self.result = self.larch_get('pca_result')
        ncomps = len(r.components)
        wmin = form['weight_min']
        if self.wids['weight_auto'].GetValue():
            nsig = r.nsig
            wmin = r.variances[nsig-1]
            self.wids['weight_min'].SetValue(wmin)
        else:
            nsig = np.where(r.variances < wmin)[0][0] - 1


        status = " Model built, %d of %d components have weight > %.4f"
        self.wids['status'].SetLabel(status %  (nsig, ncomps, wmin))
        self.wids['max_components'].SetValue(nsig+1)

        for b in ('fit_group',):
            self.wids[b].Enable()

        grid_data = []
        ind = [i for i in r.ind]
        ind.extend([0,0,0])
        ind_best = ind[r.nsig]
        for i, var in enumerate(r.variances):
            grid_data.append([var, ind[i+1], ind[i+1]/ind_best,  '0.0'])
        self.wids['table'].table.data = grid_data
        self.wids['table'].table.View.Refresh()

        self.plot_pca_components()
        self.plot_pca_weights()

    def onSavePCAModel(self, event=None):
        form = self.read_form()
        print("SAVE model: form: ", form)

    def onLoadPCAModel(self, event=None):
        form = self.read_form()



#!/usr/bin/env python
# XAS spectral decovolution
#

import numpy as np
from scipy.signal import deconvolve
from larch import parse_group_args

from larch.math import (gaussian, lorentzian, interp,
                        index_of, index_nearest, remove_dups,
                        savitzky_golay)

from .xafsutils import set_xafsGroup

def xas_deconvolve(energy, norm=None, group=None, form='lorentzian',
                   esigma=1.0, eshift=0.0, smooth=True,
                   sgwindow=None, sgorder=3, _larch=None):
    """XAS spectral deconvolution

    de-convolve a normalized mu(E) spectra with a peak shape, enhancing the
    intensity and separation of peaks of a XANES spectrum.

    The results can be unstable, and noisy, and should be used
    with caution!

    Arguments
    ----------
    energy:   array of x-ray energies (in eV) or XAFS data group
    norm:     array of normalized mu(E)
    group:    output group
    form:     functional form of deconvolution function. One of
              'gaussian' or 'lorentzian' [default]
    esigma    energy sigma to pass to gaussian() or lorentzian()
              [in eV, default=1.0]
    eshift    energy shift to apply to result. [in eV, default=0]
    smooth    whether to smooth result with savitzky_golay method [True]
    sgwindow  window size for savitzky_golay [found from data step and esigma]
    sgorder   order for savitzky_golay [3]

    Returns
    -------
    None
       The array 'deconv' will be written to the output group.

    Notes
    -----
       Support See First Argument Group convention, requiring group
       members 'energy' and 'norm'

       Smoothing with savitzky_golay() requires a window and order.  By
       default, window = int(esigma / estep) where estep is step size for
       the gridded data, approximately the finest energy step in the data.
    """
    energy, mu, group = parse_group_args(energy, members=('energy', 'norm'),
                                         defaults=(norm,), group=group,
                                         fcn_name='xas_deconvolve')
    eshift = eshift + 0.5 * esigma

    en  = remove_dups(energy)
    estep1 = int(0.1*en[0]) * 2.e-5
    en  = en - en[0]
    estep = max(estep1, 0.01*int(min(en[1:]-en[:-1])*100.0))
    npts = 1  + int(max(en) / estep)
    if npts > 25000:
        npts = 25001
        estep = max(en)/25000.0

    x = np.arange(npts)*estep
    y = interp(en, mu, x, kind='cubic')

    kernel = lorentzian
    if form.lower().startswith('g'):
        kernel = gaussian

    yext = np.concatenate((y, np.arange(len(y))*y[-1]))

    ret, err = deconvolve(yext, kernel(x, center=0, sigma=esigma))
    nret = min(len(x), len(ret))

    ret = ret[:nret]*yext[nret-1]/ret[nret-1]
    if smooth:
        if sgwindow is None:
            sgwindow = int(1.0*esigma/estep)

        sqwindow = int(sgwindow)
        if sgwindow < (sgorder+1):
            sgwindow = sgorder + 2
        if sgwindow % 2 == 0:
            sgwindow += 1
        ret = savitzky_golay(ret, sgwindow, sgorder)

    out = interp(x+eshift, ret, en, kind='cubic')
    group = set_xafsGroup(group, _larch=_larch)
    group.deconv = out


def xas_convolve(energy, norm=None, group=None, form='lorentzian',
                   esigma=1.0, eshift=0.0, _larch=None):
    """
    convolve a normalized mu(E) spectra with a Lorentzian or Gaussian peak
    shape, degrading separation of XANES features.

    This is provided as a complement to xas_deconvolve, and to deliberately
    broaden spectra to compare with spectra measured at lower resolution.

    Arguments
    ----------
    energy:   array of x-ray energies (in eV) or XAFS data group
    norm:     array of normalized mu(E)
    group:    output group
    form:     form of deconvolution function. One of
              'lorentzian' or  'gaussian' ['lorentzian']
    esigma    energy sigma (in eV) to pass to gaussian() or lorentzian() [1.0]
    eshift    energy shift (in eV) to apply to result [0]

    Returns
    -------
    None
       The array 'conv' will be written to the output group.

    Notes
    -----
       Follows the First Argument Group convention, using group members named
       'energy' and 'norm'
    """

    energy, mu, group = parse_group_args(energy, members=('energy', 'norm'),
                                         defaults=(norm,), group=group,
                                         fcn_name='xas_convolve')
    eshift = eshift + 0.5 * esigma

    en  = remove_dups(energy)
    en  = en - en[0]
    estep = max(0.001, 0.001*int(min(en[1:]-en[:-1])*1000.0))

    npad = 1 + int(max(estep*2.01, 50*esigma)/estep)

    npts = npad  + int(max(en) / estep)

    x = np.arange(npts)*estep
    y = interp(en, mu, x, kind='cubic')

    kernel = lorentzian
    if form.lower().startswith('g'):
        kernel = gaussian

    k = kernel(x, center=0, sigma=esigma)
    ret = np.convolve(y, k, mode='full')

    out = interp(x-eshift, ret[:len(x)], en, kind='cubic')

    group = set_xafsGroup(group, _larch=_larch)
    group.conv = out / k.sum()



#!/usr/bin/env python
"""
utilities for larch
"""
from __future__ import print_function
import re
import sys
from base64 import b64encode, b32encode
import hashlib
import random
from distutils.version import StrictVersion

def bytes2str(s):
    if isinstance(s, str):
        return s
    elif isinstance(s, bytes):
        return s.decode(sys.stdout.encoding)
    return str(s, sys.stdout.encoding)
def str2bytes(s):
    'string to byte conversion'
    if isinstance(s, bytes):
        return s
    return bytes(s, sys.stdout.encoding)


RESERVED_WORDS = ('and', 'as', 'assert', 'break', 'class', 'continue',
                  'def', 'del', 'elif', 'else', 'eval', 'except', 'exec',
                  'execfile', 'finally', 'for', 'from', 'global', 'if',
                  'import', 'in', 'is', 'lambda', 'not', 'or', 'pass',
                  'print', 'raise', 'return', 'try', 'while', 'with',
                  'group', 'end', 'endwhile', 'endif', 'endfor', 'endtry',
                  'enddef', 'True', 'False', 'None')

NAME_MATCH = re.compile(r"[a-zA-Z_][a-zA-Z0-9_]*(\.[a-zA-Z_][a-zA-Z0-9_]*)*$").match
VALID_SNAME_CHARS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_'
VALID_NAME_CHARS = '.%s' % VALID_SNAME_CHARS
VALID_CHARS1 = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_'

BAD_FILECHARS = ';~,`!%$@$&^?*#:"/|\'\\\t\r\n (){}[]<>'
GOOD_FILECHARS = '_'*len(BAD_FILECHARS)

BAD_VARSCHARS = BAD_FILECHARS + '=+-.'
GOOD_VARSCHARS = '_'*len(BAD_VARSCHARS)

TRANS_FILE = str.maketrans(BAD_FILECHARS, GOOD_FILECHARS)
TRANS_VARS = str.maketrans(BAD_VARSCHARS, GOOD_VARSCHARS)


def PrintExceptErr(err_str, print_trace=True):
    " print error on exceptions"
    print('\n***********************************')
    print(err_str)
    #print 'PrintExceptErr', err_str
    try:
        print('Error: %s' % sys.exc_type)
        etype, evalue, tback = sys.exc_info()
        if print_trace == False:
            tback = ''
        sys.excepthook(etype, evalue, tback)
    except:
        print('Error printing exception error!!')
        raise
    print('***********************************\n')

def strip_comments(sinp, char='#'):
    "find character in a string, skipping over quoted text"
    if sinp.find(char) < 0:
        return sinp
    i = 0
    while i < len(sinp):
        tchar = sinp[i]
        if tchar in ('"',"'"):
            eoc = sinp[i+1:].find(tchar)
            if eoc > 0:
                i = i + eoc
        elif tchar == char:
            return sinp[:i].rstrip()
        i = i + 1
    return sinp

def strip_quotes(t):
    d3, s3, d1, s1 = '"""', "'''", '"', "'"
    if hasattr(t, 'startswith'):
        if ((t.startswith(d3) and t.endswith(d3)) or
            (t.startswith(s3) and t.endswith(s3))):
            t = t[3:-3]
        elif ((t.startswith(d1) and t.endswith(d1)) or
              (t.startswith(s1) and t.endswith(s1))):
            t = t[1:-1]
    return t

def isValidName(name):
    "input is a valid name"
    if name in RESERVED_WORDS:
        return False
    tnam = name[:].lower()
    return NAME_MATCH(tnam) is not None

def fixName(name, allow_dot=True):
    "try to fix string to be a valid name"
    if isValidName(name):
        return name

    if isValidName('_%s' % name):
        return '_%s' % name
    chars = []
    valid_chars = VALID_SNAME_CHARS
    if allow_dot:
        valid_chars = VALID_NAME_CHARS
    for s in name:
        if s not in valid_chars:
            s = '_'
        chars.append(s)
    name = ''.join(chars)
    # last check (name may begin with a number or .)
    if not isValidName(name):
        name = '_%s' % name
    return name


def fix_filename(s):
    """fix string to be a 'good' filename.
    This may be a more restrictive than the OS, but
    avoids nasty cases."""
    t = str(s).translate(TRANS_FILE)
    if t.count('.') > 1:
        for i in range(t.count('.') - 1):
            idot = t.find('.')
            t = "%s_%s" % (t[:idot], t[idot+1:])
    return t

def fix_varname(s):
    """fix string to be a 'good' variable name."""
    t = str(s).translate(TRANS_VARS)

    if len(t) < 1:
        t = '_unlabeled_'
    if t[0] not in VALID_CHARS1:
        t = '_%s' % t
    while t.endswith('_'):
        t = t[:-1]
    return t

def common_startstring(words):
    """common starting substring for a list of words"""
    out = words[0]
    for tmp in words[1:]:
        i = 0
        for a, b in zip(out, tmp):
            if a == b:
                i += 1
            else:
                out = out[:i]
    return out


def unique_name(name, nlist, max=1000):
    """return name so that is is not in list,
    by appending _1, _2, ... as necessary up to a max suffix

    >>> unique_name('foo', ['bar, 'baz'])
    'foo'

    >>> unique_name('foo', ['foo', 'bar, 'baz'])
    'foo_1'

    """
    out = name
    if name in nlist:
        for i in range(1, max+1):
            out = "%s_%i"  % (name, i)
            if out not in nlist:
                break
    return out


def isNumber(num):
    "input is a number"
    try:
        cnum = complex(num)
        return True
    except ValueError:
        return False

def isLiteralStr(inp):
    "is a literal string"
    return ((inp.startswith("'") and inp.endswith("'")) or
            (inp.startswith('"') and inp.endswith('"')))


def find_delims(s, delim='"',match=None):
    """find matching delimeters (quotes, braces, etc) in a string.
    returns
      True, index1, index2 if a match is found
      False, index1, len(s) if a match is not found
    the delimiter can be set with the keyword arg delim,
    and the matching delimiter with keyword arg match.

    if match is None (default), match is set to delim.

    >>> find_delims(mystr, delim=":")
    >>> find_delims(mystr, delim='<', match='>')
    """
    esc, dbesc = "\\", "\\\\"
    if match is None:
        match = delim
    j = s.find(delim)
    if j > -1 and s[j:j+len(delim)] == delim:
        p1, p2, k = None, None, j
        while k < j+len(s[j+1:]):
            k = k+1
            if k > 0: p1 = s[k-1:k]
            if k > 1: p2 = s[k-2:k]
            if (s[k:k+len(match)] == match and not (p1 == esc and p2 != dbesc)):
                return True, j, k+len(match)-1
            p1 = s[k:k+1]
    return False, j, len(s)

def version_ge(v1, v2):
    "returns whether version string 1 >= version_string2"
    return StrictVersion(bytes2str(v1)) >= StrictVersion(bytes2str(v2))


def b32hash(s):
    """return a base32 hash of a string"""
    _hash = hashlib.sha256()
    _hash.update(str2bytes(s))
    return bytes2str(b32encode(_hash.digest()))

def b64hash(s):
    """return a base64 hash of a string"""
    _hash = hashlib.sha256()
    _hash.update(str2bytes(s))
    return bytes2str(b64encode(_hash.digest()))

def file2groupname(filename, slen=5, symtable=None):
    """create a group name based of filename
    the group name will have a string component of
    length slen followed by a 4 digit number

    Arguments
    ---------
    filename  (str)  filename to use
    slen      (int)  length of string portion (default 4)
    symtable  (None or larch symbol table) symbol table for
              checking that the group name is unique
    """
    def randstr(n):
        return ''.join([chr(random.randint(97, 122)) for i in range(n)])

    gname = fix_varname(filename).lower() +  randstr(slen)
    if '_' in gname:
        gname = gname.replace('_', '')
        gname = fix_varname(gname)

    fmt, count, maxcount = "%s{:04d}", 1, 999
    fstr = fmt % (gname[:slen])
    gname = fstr.format(count)
    if symtable is not None:
        scount = 0
        while hasattr(symtable, gname):
            count += 1
            if count > maxcount:
                scount += 1
                count = 1
                fstr = fmt % randstr(slen)
            gname = fstr.format(count)
            if scount > 1000:
                raise ValueError("exhausted unique group names")
    return gname



import wx
from wx.richtext import RichTextCtrl

from wxutils import (pack, FRAMESTYLE, is_wxPhoenix)

class ReportFrame(wx.Frame):
    """basic frame for displaying a text report -- should be improved!
    """
    def __init__(self, parent=None, text=None, size=(550, 600), **kws):
        wx.Frame.__init__(self, parent, size=size, style=FRAMESTYLE)
        self.report = RichTextCtrl(self,size=(500, 500),
                                   style=wx.VSCROLL)

        self.report.SetEditable(False)
        self.report.SetFont(Font(11))
        self.reportx.SetMinSize((500, 500))

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.report, 1, wx.ALL|wx.GROW, 2)
        pack(self, sizer)
        if text is not None:
            self.set_text(text)
        self.Show()
        self.Raise()

    def set_text(self, text):
        self.report.SetEditable(True)
        self.report.SetValue(text)
        self.report.SetEditable(False)



#!/usr/bin/env python
"""
Code to write and read CVS files

"""
import sys
import os
import time
import json
import platform
import csv
from collections import OrderedDict

import numpy as np
from dateutil.parser import parse as dateparse
from larch import Group
from larch.math import interp
from larch.utils.strutils import bytes2str, fix_varname

maketrans = str.maketrans

def groups2csv(grouplist, filename,
               x='energy', y='norm', _larch=None):
    """save data from a list of groups to a CSV file

    Arguments
    ---------
    grouplist  list of groups to save arrays from
    filname    name of output file
    x          name of group member to use for `x`
    y          name of group member to use for `y`

    """
    def get_label(grp):
        'get label for group'
        for attr in ('filename', 'label', 'name', 'file', '__name__'):
            o = getattr(grp, attr, None)
            if o is not None:
                return o
        return repr(o)

    ngroups = len(grouplist)
    x0 = getattr(grouplist[0], x)
    npts = len(x0)
    columns = [x0, getattr(grouplist[0], y)]
    labels = [x, get_label(grouplist[0]) ]

    for g in grouplist[1:]:

        labels.append(get_label(g))
        _x = getattr(g, x)
        _y = getattr(g, y)

        if ((len(_x) != npts) or (abs(_x -x0)).sum() > 1.0):
            columns.append(interp(_x, _y, x0))
        else:
            columns.append(_y)

    buff = ["# %s" % ', '.join(labels)]
    for i in range(npts):
        buff.append(', '.join(["%.6f" % s[i] for s in columns]))

    buff.append('')
    with open(filename, 'w') as fh:
        fh.write("\n".join(buff))

    print("Wrote %i groups to %s" % (len(columns)-1, filename))


def str2float(word, allow_times=True):
    """convert a work to a float

    Arguments
    ---------
      word          str, word to be converted
      allow_times   bool, whether to support time stamps [True]

    Returns
    -------
      either a float or text

    Notes
    -----
      The `allow_times` will try to support common date-time strings
      using the dateutil module, returning a numerical value as the
      Unix timestamp, using
          time.mktime(dateutil.parser.parse(word).timetuple())
    """
    mktime = time.mktime
    val = word
    try:
        val = float(word)
    except ValueError:
        try:
            val = mktime(dateparse(word).timetuple())
        except ValueError:
            pass
    return val

def read_csv(filename):
    """read CSV file, return group with data as columns"""
    csvfile = open(filename, 'r')
    dialect = csv.Sniffer().sniff(csvfile.read(),  [',',';', '\t'])
    csvfile.seek(0)

    data = None
    isfloat = None
    for row in csv.reader(csvfile, dialect):
        if data is None:
            ncols = len(row)
            data = [[] for i in range(ncols)]
            isfloat =[None]*ncols
        for i, word in enumerate(row):
            data[i].append(str2float(word))
            if isfloat[i] is None:
                try:
                    _ = float(word)
                    isfloat[i] = True
                except ValueError:
                    isfloat[i] = False

    out = Group(filename=filename, data=data)
    for icol in range(ncols):
        cname = 'col_%2.2d' % (icol+1)
        val = data[icol]
        if isfloat[icol]:
            val = np.array(val)
        setattr(out, cname, val)

    return out



import sys
import os
from os.path import realpath, isdir, isfile, join, basename, dirname, abspath
import glob
from shutil import copy, move
import subprocess
import time
import re
from optparse import OptionParser
from subprocess import Popen, PIPE

from larch import Group, isNamedClass
from larch.utils import isotime, bytes2str, uname, bindir

def find_exe(exename):
    if uname == 'win' and not exename.endswith('.exe'):
        exename = "%s.exe" % exename
    exefile = join(bindir, exename)
    if os.path.exists(exefile) and os.access(exefile, os.X_OK):
        return exefile

class FeffRunner(Group):
    """
    A Larch plugin for managing calls to the feff85exafs stand-alone executables.
    This plugin does not manage the output of Feff.  See feffpath() and other tools.

    Methods:
        run -- run one or more parts of feff

            feff = feffrunner(feffinp='path/to/feff.inp')
            feff.run() to run feff monolithically
            feff.run('rdinp')
            feff.run('xsph')
               and so on to run individual parts of feff
               ('rdinp', 'pot', 'opconsat', 'xsph', 'pathfinder', 'genfmt', 'ff2x')

          If the symbol _xafs._feff_executable is set to a Feff executable,
          it can be run by doing

            feff = feffrunner(feffinp='path/to/feff6.inp')
            feff.run(None)

          run returns None if feff ran successfully, otherwise it
          returns an Exception with a useful message

          Other versions of feff in the execution path can also be
          handled, with the caveat that the executable begins with
          'feff', i.e. 'feff6', 'feff7', etc.

            feff = feffrunner(feffinp='path/to/feff6.inp')
            feff.run('feff6')

          If the value of the feffinp attribute is a file with a
          basename other than 'feff.inp', that file will be renamed to
          'feff.inp' and care will be taken to preserve an existing
          file by that name.

    Attributes:
        folder   -- the folder to run in, containing feff.inp file
        feffinp  -- the feff.inp file, absolute or relative to `folder`
        resolved -- the fully resolved path to the most recently run executable
        verbose  -- write screen messages if True
        mpse     -- run opconsat after pot if True

    """

    Feff8l_modules = ('rdinp', 'pot', 'xsph', 'pathfinder', 'genfmt', 'ff2x')

    def __init__(self, feffinp='feff.inp', folder='.', verbose=True, _larch=None, **kws):
        kwargs = dict(name='Feff runner')
        kwargs.update(kws)
        Group.__init__(self,  **kwargs)
        self._larch = _larch

        if folder is None:
            folder = '.'
        self.folder   = folder
        self.feffinp  = feffinp
        self.verbose  = verbose
        self.mpse     = False
        self.resolved = None
        self.threshold = []
        self.chargetransfer = []

    def __repr__(self):
        fullfile = os.path.join(self.folder, self.feffinp)
        return '<External Feff Group: %s>' % fullfile

    def run(self, feffinp=None, folder=None, exe='feff8l'):
        """
        Make system call to run one or more of the stand-alone executables,
        writing a log file to the folder containing the input file.

        """
        if folder is not None:
            self.folder = folder

        if feffinp is not None:
            self.feffinp = feffinp

        if self.feffinp is None:
            raise Exception("no feff.inp file was specified")

        savefile = '.save_.inp'
        here = abspath(os.getcwd())
        os.chdir(abspath(self.folder))

        feffinp_dir, feffinp_file = os.path.split(self.feffinp)
        feffinp_dir = dirname(self.feffinp)
        if len(feffinp_dir) > 0:
            os.chdir(feffinp_dir)

        if not isfile(feffinp_file):
            raise Exception("feff.inp file '%s' could not be found" % feffinp_file)

        if exe in (None, 'feff8l'):
            for module in self.Feff8l_modules:
                os.chdir(here)
                self.run(exe=module)
            return

        #
        # exe is set, find the corresponding executable file
        ## find program to run:
        program = None
        if exe in self.Feff8l_modules:
            exe = "feff8l_%s" % exe

        resolved_exe = find_exe(exe)
        if resolved_exe is not None:
            program = resolved_exe

        else:
            try:
                program = self._larch.symtable.get_symbol('_xafs._feff_executable')
            except (NameError, AttributeError) as exc:
                program = None

        if program is not None:
            if not os.access(program, os.X_OK):
                program = None

        if program is None:  # Give up!
            os.chdir(here)
            raise Exception("'%s' executable cannot be found" % exe)

        ## preserve an existing feff.inp file if this is not called feff.inp
        if feffinp_file != 'feff.inp':
            if isfile('feff.inp'):
                copy('feff.inp', savefile)
            copy(feffinp_file, 'feff.inp')

        _, logname = os.path.split(program)
        if logname.endswith('.exe'):
            logname = logname[:4]

        log = 'feffrun_%s.log' % logname

        if isfile(log):
            os.unlink(log)

        f = open(log, 'a')
        header = "\n======== running Feff module %s ========\n" % exe

        def write(msg):
            msg = bytes2str(msg)
            msg = " : {:s}\n".format(msg.strip().rstrip())
            if self._larch is not None:
                self._larch.writer.write(msg)
            else:
                sys.stdout.write(msg)

        if self.verbose:
            write(header)
        f.write(header)
        process=subprocess.Popen(program, shell=False,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.STDOUT)
        flag = False
        thislist = []
        while True:
            if  process.returncode is None:
                process.poll()
            time.sleep(0.01)
            line = bytes2str(process.stdout.readline())
            if not line:
                break
            if self.verbose:
                write(line)
            ## snarf threshold energy
            pattern = re.compile('mu_(new|old)=\s+(-?\d\.\d+)')
            match = pattern.search(line)
            if match is not None:
                self.threshold.append(match.group(2))
            ## snarf charge transfer
            if line.strip().startswith('Charge transfer'):
                thislist = []
                flag = True
            elif line.strip().startswith('SCF ITERATION'):
                self.chargetransfer.append(list(thislist))
                flag = False
            elif line.strip().startswith('Done with module 1'):
                self.chargetransfer.append(list(thislist))
                flag = False
            elif flag:
                this = line.split()
                thislist.append(this[1])
            f.write(line)
        f.close

        if isfile(savefile):
            move(savefile, 'feff.inp')
        os.chdir(here)
        return None

######################################################################
def feffrunner(folder=None, feffinp=None, verbose=True, _larch=None, **kws):
    """
    Make a FeffRunner group given a folder containing a baseline calculation
    """
    return FeffRunner(folder=folder, feffinp=feffinp, verbose=verbose,
                      _larch=_larch, **kws)

def feff6l(feffinp='feff.inp', folder='.', verbose=True, _larch=None, **kws):
    """
    run a Feff6l calculation for a feff.inp file in a folder

    Arguments:
    ----------
      feffinp (str): name of feff.inp file to use ['feff.inp']
      folder (str): folder for calculation, containing 'feff.inp' file ['.']
      verbose (bool): whether to print out extra messages [False]

    Returns:
    --------
      instance of FeffRunner

    Notes:
    ------
      many results data files are generated in the Feff working folder
    """
    feffrunner = FeffRunner(folder=folder, feffinp=feffinp, verbose=verbose, _larch=_larch)
    exe = find_exe('feff6l')
    feffrunner.run(exe=exe)
    return feffrunner

def feff6l_cli():
    """run feff6l as command line program
    """

    usage = """usage: %prog [options] folder(s)

run feff6l on one or more folders containing feff.inp files
or on an input file in the current folder

Examples:
   feff6l Structure1 Structure2

   feff6l feff_Cu2O.inp

"""

    parser = OptionParser(usage=usage, prog="feff6l",
                          version="Feff6L version 6L.02")

    FEFFINP = 'feff.inp'
    (options, args) = parser.parse_args()
    if len(args) == 0:
        args = ['.']

    curdir = abspath(os.getcwd())
    for arg in args:
        if os.path.isfile(arg):
            feff6l(feffinp=arg)
        elif os.path.isdir(arg):
            feffinp = os.path.join(arg, 'feff.inp')
            if os.path.exists(feffinp):
                os.chdir(abspath(arg))
                feff6l(folder=arg)
            else:
                msg = "Could not find feff.inp file in folder '{:s}'"
                sys.stdout.write(msg.format(abspath(os.curdir)))
            os.chdir(curdir)


def feff8l(feffinp='feff.inp', folder='.', module=None, verbose=True, _larch=None, **kws):
    """
    run a Feff8l calculation for a feff.inp file in a folder

    Arguments:
    ----------
      feffinp (str): name of feff.inp file to use ['feff.inp']
      folder (str): folder for calculation, containing 'feff.inp' file ['.']
      module (None or str): module of Feff8l to run [None -- run all]
      verbose (bool): whether to print out extra messages [False]

    Returns:
    --------
      instance of FeffRunner

    Notes:
    ------
      many results data files are generated in the Feff working folder
    """
    feffrunner = FeffRunner(folder=folder, feffinp=feffinp, verbose=verbose, _larch=_larch)
    feffrunner.run(exe='feff8l')
    return feffrunner


def feff8l_cli():
    """run feff8l as  a command line program to run all or some of
     feff8l_rdinp
     feff8l_pot
     feff8l_xsph
     feff8l_pathfinder
     feff8l_genfmt
     feff8l_ff2x
    """

    usage = """usage: %prog [options] folder(s)

run feff8l (or selected modules) on one
or more folders containing feff.inp files.

Example:
   feff8l --no-ff2chi Structure1 Structure2
"""

    parser = OptionParser(usage=usage, prog="feff8l",
                          version="Feff85L for EXAFS version 8.5L, build 001")
    parser.add_option("-q", "--quiet", dest="quiet", action="store_true",
                      default=False, help="set quiet mode, default=False")
    parser.add_option("--no-pot", dest="no_pot", action="store_true",
                      default=False, help="do not run POT module")
    parser.add_option("--no-phases", dest="no_phases", action="store_true",
                      default=False, help="do not run XSPH module")
    parser.add_option("--no-paths", dest="no_paths", action="store_true",
                      default=False, help="do not run PATHFINDER module")
    parser.add_option("--no-genfmt", dest="no_genfmt", action="store_true",
                      default=False, help="do not run GENFMT module")
    parser.add_option("--no-ff2chi", dest="no_ff2chi", action="store_true",
                      default=False, help="do not run FF2CHI module")


    FEFFINP = 'feff.inp'
    (options, args) = parser.parse_args()

    verbose = not options.quiet
    modules = ['rdinp', 'pot', 'xsph', 'pathfinder', 'genfmt', 'ff2x']
    if options.no_pot:
        modules.remove('pot')
    if options.no_phases:
        modules.remove('xsph')
    if options.no_paths:
        modules.remove('pathfinder')
    if options.no_genfmt:
        modules.remove('genfmt')
    if options.no_ff2chi:
        modules.remove('ff2x')

    if len(args) == 0:
        args = ['.']


    def run_feff8l(modules):
        """ run selected modules of Feff85L   """
        try:
            logfile = open('feff8l.log', 'w+')
        except:
            logfile = tempfile.NamedTemporaryFile(prefix='feff8l')

        def write(msg):
            msg = bytes2str(msg)
            sys.stdout.write(msg)
            logfile.write(msg)

        write("#= Feff85l %s\n" % isotime())
        for mod in modules:
            write("#= Feff85l %s module\n" % mod)
            exe = find_exe('feff8l_%s' % mod)
            proc = Popen(exe, stdout=PIPE, stderr=PIPE)
            while True:
                msg = bytes2str(proc.stdout.read())
                if msg == '':
                    break
                write(msg)
            while True:
                msg = bytes2str(proc.stderr.read())
                if msg == '':
                    break
                write("#ERROR %s" % msg)
            logfile.flush()
        for fname in glob.glob('log*.dat'):
            try:
                os.unlink(fname)
            except IOError:
                pass
        write("#= Feff85l done %s\n" % isotime())

    for dirname in args:
        if os.path.exists(dirname) and os.path.isdir(dirname):
            thisdir = abspath(os.curdir)
            os.chdir(dirname)
            if os.path.exists(FEFFINP) and os.path.isfile(FEFFINP):
                run_feff8l(modules)
            else:
                msg = "Could not find feff.inp file in folder '{:s}'"
                sys.stdout.write(msg.format(abspath(os.curdir)))
            os.chdir(thisdir)
        else:
            print("Could not find folder '{:s}'".format(dirname))



#!/usr/bin/env python
"""
GUI for displaying maps from HDF5 files

"""

import os
import platform
import sys
import time
import json
import socket
import datetime
from functools import partial
from threading import Thread

import wx
import wx.lib.scrolledpanel as scrolled
import wx.lib.mixins.inspection
try:
    from wx._core import PyDeadObjectError
except:
    PyDeadObjectError = Exception


import h5py
import numpy as np

from ..wxlib import (LarchPanel, LarchFrame, EditableListBox, SimpleText,
                     FloatCtrl, Font, pack, Popup, Button, MenuItem,
                     Choice, Check, GridPanel, FileSave, HLine)
from ..utils.strutils import bytes2str, version_ge

from ..xrmmap import GSEXRM_MapFile, GSEXRM_FileStatus, h5str, ensure_subgroup

CEN = wx.ALIGN_CENTER
LEFT = wx.ALIGN_LEFT
RIGHT = wx.ALIGN_RIGHT
ALL_CEN =  wx.ALL|CEN
ALL_LEFT =  wx.ALL|LEFT

VARNAMES = ('a', 'b', 'c', 'd', 'e', 'f')

class MapMathPanel(scrolled.ScrolledPanel):
    """Panel of Controls for doing math on arrays from Map data"""
    label  = 'Map Math'
    def __init__(self, parent, owner=None, **kws):

        self.map   = None
        self.cfile = None

        scrolled.ScrolledPanel.__init__(self, parent, -1,
                                        style=wx.GROW|wx.TAB_TRAVERSAL, **kws)
        self.owner = owner
        sizer = wx.GridBagSizer(3, 3)
        bpanel = wx.Panel(self)
        show_new = Button(bpanel, 'Show New Map',     size=(125, -1),
                          action=partial(self.onShowMap, new=True))
        show_old = Button(bpanel, 'Replace Last Map', size=(125, -1),
                                   action=partial(self.onShowMap, new=False))
        bsizer = wx.BoxSizer(wx.HORIZONTAL)
        bsizer.Add(show_new, 0, 3)
        bsizer.Add(show_old, 0, 3)
        pack(bpanel, bsizer)

        save_arr = Button(self, 'Save Array', size=(120, -1),
                          action=self.onSaveArray)

        self.expr_in = wx.TextCtrl(self, -1,   '', size=(180, -1))
        self.name_in = wx.TextCtrl(self, -1,   '', size=(180, -1))

        ir = 0
        txt = """Enter Math Expressions for Map: a+b, (a-b)/c, log10(a+0.1),  etc"""
        sizer.Add(SimpleText(self, txt),    (ir, 0), (1, 6), ALL_LEFT, 2)

        ir += 1
        sizer.Add(SimpleText(self, 'Expression:'),    (ir, 0), (1, 1), ALL_CEN, 2)
        sizer.Add(self.expr_in,   (ir, 1), (1, 1), ALL_LEFT, 2)
        sizer.Add(bpanel,  (ir, 2), (1, 3), ALL_LEFT, 2)

        ir += 1
        sizer.Add(SimpleText(self, 'Array Name:'),    (ir, 0), (1, 1), ALL_CEN, 2)
        sizer.Add(self.name_in,   (ir, 1), (1, 1), ALL_LEFT, 2)
        sizer.Add(save_arr,  (ir, 2), (1, 1), ALL_LEFT, 2)

        ir += 1
        sizer.Add(SimpleText(self, 'Array'),       (ir, 0), (1, 1), ALL_CEN, 2)
        sizer.Add(SimpleText(self, 'File'),        (ir, 1), (1, 1), ALL_CEN, 2)
        sizer.Add(SimpleText(self, 'Detector'),    (ir, 2), (1, 1), ALL_CEN, 2)
        sizer.Add(SimpleText(self, 'ROI'),         (ir, 3), (1, 1), ALL_CEN, 2)
        sizer.Add(SimpleText(self, 'DT Correct?'), (ir, 4), (1, 1), ALL_CEN, 2)
        sizer.Add(SimpleText(self, 'Array Shape'), (ir, 5), (1, 1), ALL_CEN, 2)

        self.varfile  = {}
        self.varroi   = {}
        self.varshape = {}
        self.vardet   = {}
        self.varcor   = {}
        for varname in VARNAMES:
            self.varfile[varname]  = vfile  = Choice(self, size=(250, -1),
                                                     action=partial(self.onFILE, varname=varname))
            self.varroi[varname]   = vroi   = Choice(self, size=(125, -1),
                                                     action=partial(self.onROI, varname=varname))
            self.vardet[varname]   = vdet   = Choice(self, size=(100, -1),
                                                     action=partial(self.onDET, varname=varname))
            self.varcor[varname]   = vcor   = wx.CheckBox(self, -1, ' ')
            self.varshape[varname] = vshape = SimpleText(self, '(, )',
                                                          size=(125, -1))
            vcor.SetValue(self.owner.dtcor)
            vdet.SetSelection(0)

            ir += 1
            sizer.Add(SimpleText(self, '%s = ' % varname),    (ir, 0), (1, 1), ALL_CEN, 2)
            sizer.Add(vfile,                        (ir, 1), (1, 1), ALL_CEN, 2)
            sizer.Add(vdet,                         (ir, 2), (1, 1), ALL_CEN, 2)
            sizer.Add(vroi,                         (ir, 3), (1, 1), ALL_CEN, 2)
            sizer.Add(vcor,                         (ir, 4), (1, 1), ALL_CEN, 2)
            sizer.Add(vshape,                       (ir, 5), (1, 1), ALL_LEFT, 2)

        ir += 1
        sizer.Add(HLine(self, size=(350, 4)), (ir, 0), (1, 5), ALL_LEFT, 2)

        ir += 1
        sizer.Add(SimpleText(self, 'Work Arrays: '), (ir, 0), (1, 1), ALL_LEFT, 2)

        self.workarray_choice = Choice(self, size=(200, -1),
                                       action=self.onSelectArray)
        btn_delete  = Button(self, 'Delete Array',  size=(90, -1),
                              action=self.onDeleteArray)
        self.info1   = wx.StaticText(self, -1, '',  size=(250, -1))
        self.info2   = wx.StaticText(self, -1, '',  size=(250, -1))


        sizer.Add(self.workarray_choice, (ir, 1), (1, 1), ALL_LEFT, 2)
        sizer.Add(btn_delete, (ir, 2), (1, 1), ALL_LEFT, 2)
        sizer.Add(self.info1, (ir+1, 0), (1, 3), ALL_LEFT, 2)
        sizer.Add(self.info2, (ir+2, 0), (3, 3), ALL_LEFT, 2)

        pack(self, sizer)
        self.SetupScrolling()

    def on_select(self):
        for v in VARNAMES:
            self.onFILE(evt=None, varname=v)


    def onSelectArray(self, evt=None):
        xrmfile = self.owner.current_file
        name = self.workarray_choice.GetStringSelection()
        dset = xrmfile.get_work_array(h5str(name))
        expr = bytes2str(dset.attrs.get('expression', '<unknonwn>'))
        self.info1.SetLabel("Expression: %s" % expr)

        info = json.loads(bytes2str(dset.attrs.get('info', [])))
        buff = []
        for var, dat in info:
            fname, aname, det, dtc = dat
            if dat[1] != '1':
                buff.append("%s= %s('%s', det=%s, dtcorr=%s)" % (var, fname, aname, det, dtc))
        self.info2.SetLabel('\n'.join(buff))

    def onDeleteArray(self, evt=None):
        name = self.workarray_choice.GetStringSelection()
        xrmfile = self.owner.current_file

        if (wx.ID_YES == Popup(self.owner, """Delete Array '%s' for %s?
    WARNING: This cannot be undone
    """ % (name, xrmfile.filename),
                               'Delete Array?', style=wx.YES_NO)):
                xrmfile.del_work_array(h5str(name))
                self.update_xrmmap(xrmfile)

    def onSaveArray(self, evt=None):
        name = self.name_in.GetValue()
        expr = self.expr_in.GetValue()
        xrmfile = self.owner.current_file
        info = []
        for varname in sorted(self.varfile.keys()):
            fname   = self.varfile[varname].GetStringSelection()
            roiname = self.varroi[varname].GetStringSelection()
            dname   = self.vardet[varname].GetStringSelection()
            dtcorr  = self.varcor[varname].IsChecked()
            info.append((varname, (fname, roiname, dname, dtcorr)))

        if self.map is None:
            self.onShowMap()

        if name in xrmfile.work_array_names():
            if (wx.ID_YES == Popup(self.owner, """Overwrite Array '%s' for %s?
    WARNING: This cannot be undone
    """ % (name, xrmfile.filename),
                                   'Overwrite Array?', style=wx.YES_NO)):
                xrmfile.del_work_array(h5str(name))
            else:
                return
        xrmfile.add_work_array(self.map, h5str(name),
                               expression=h5str(expr),
                               info=json.dumps(info))

        for p in self.owner.nb.pagelist:
            if hasattr(p, 'update_xrmmap'):
                p.update_xrmmap(xrmfile=xrmfile, set_detectors=True)

    def onFILE(self, evt=None, varname='a'):

        fname   = self.varfile[varname].GetStringSelection()
        if fname not in (None, 'None', ''):
            shape = self.owner.filemap[fname].get_shape()
            self.varshape[varname].SetLabel('%s' % repr(shape))

    def onDET(self, evt, varname='a'):
        self.set_roi_choices(varname=varname)

    def onROI(self, evt, varname='a'):
        fname   = self.varfile[varname].GetStringSelection()
        roiname = self.varroi[varname].GetStringSelection()
        dname   = self.vardet[varname].GetStringSelection()
        dtcorr  = self.varcor[varname].IsChecked()

        shape = self.owner.filemap[fname].get_shape()
        self.varshape[varname].SetLabel('%s' % repr(shape))
        map = self.owner.filemap[fname].get_roimap(roiname, det=dname, dtcorrect=dtcorr)

    def update_xrmmap(self, xrmfile=None, set_detectors=None):
        if xrmfile is None:
            xrmfile = self.owner.current_file

        self.cfile = xrmfile
        self.xrmmap = xrmfile.xrmmap
        self.set_file_choices(self.owner.filelist.GetItems())
        self.set_det_choices()
        self.set_workarray_choices(self.xrmmap)

        for vfile in self.varfile.values():
            vfile.SetSelection(-1)
        self.info1.SetLabel('')
        self.info2.SetLabel('')

    def set_det_choices(self, varname=None):

        det_list = self.cfile.get_detector_list()
        if varname is None:
            for wid in self.vardet.values():
                wid.SetChoices(det_list)
        else:
            self.vardet[varname].SetChoices(det_list)
        self.set_roi_choices(varname=varname)


    def set_roi_choices(self, varname=None):
        if varname is None:
            for varname in self.vardet.keys():
                dname = self.vardet[varname].GetStringSelection()
                rois = self.update_roi(dname)
                self.varroi[varname].SetChoices(rois)
        else:
            dname = self.vardet[varname].GetStringSelection()
            rois = self.update_roi(dname)
            self.varroi[varname].SetChoices(rois)


    def update_roi(self, detname):
        return self.cfile.get_roi_list(detname)

    def set_workarray_choices(self, xrmmap):

        c = self.workarray_choice
        c.Clear()
        if 'work' in xrmmap:
            choices = [a for a in xrmmap['work']]
            c.AppendItems(choices)
            c.SetSelection(len(choices)-1)

    def set_file_choices(self, fnames):
        for wid in self.varfile.values():
            wid.SetChoices(fnames)

        for varname in VARNAMES:
            fname = self.varfile[varname].GetStringSelection()
            if fname in (None, 'None', ''):
                self.varfile[varname].SetSelection(0)


    def onShowMap(self, event=None, new=True):
        def get_expr(wid):
            val = str(wid.Value)
            if len(val) == 0:
                val = '1'
            return val
        expr_in = get_expr(self.expr_in)

        files = []
        _larch = self.owner.larch
        filemap = self.owner.filemap

        for varname in self.varfile.keys():
            try:
                _larch.symtable.get_symbol(varname)
                _larch.symtable.del_symbol(str(varname))
            except:
                pass
            fname   = self.varfile[varname].GetStringSelection()
            if fname in filemap:
                if fname not in files:
                    files.append(fname)
                roiname = self.varroi[varname].GetStringSelection()
                dname   = self.vardet[varname].GetStringSelection()
                dtcorr  = self.varcor[varname].IsChecked()

                map = filemap[fname].get_roimap(roiname, det=dname, dtcorrect=dtcorr)
                _larch.symtable.set_symbol(str(varname), map)

        _larch.eval("_mapcalc = %s" % expr_in)
        self.map = _larch.symtable._mapcalc
        omap = self.map[:, 1:-1]
        info  = 'Intensity: [%g, %g]' %(omap.min(), omap.max())
        title = '%se: %s' % (fname, expr_in)
        subtitles = None

        main_file = filemap[files[0]]
        try:
            x = main_file.get_pos(0, mean=True)
        except:
            x = None
        try:
            y = main_file.get_pos(1, mean=True)
        except:
            y = None

        if len(self.owner.im_displays) == 0 or new:
            iframe = self.owner.add_imdisplay(title)

        self.owner.display_map(omap, title=title, subtitles=subtitles,
                               info=info, x=x, y=y, xrmfile=main_file,
                               with_savepos=(len(files)==1))



"""
Methods for fitting background in xray spectra (energy dispersive or diffraction)

Authors/Modifications:
----------------------
* Mark Rivers, GSECARS
* modified for Larch, M Newville
* modified to work for XRD data, M. Koker

Notes:
------
This function fits a background to MCA spectrum or XRD data. The background is
fitted using an enhanced version of the algorithm published by
Kajfosz, J. and Kwiatek, W .M. (1987)  'Non-polynomial approximation of
background in x-ray spectra.' Nucl. Instrum. Methods B22, 78-81.


Procedure:

1) A series of concave down polynomials are constructed. At each
channel "i" an n'th degree polynomial which is concave up is
fitted. The polynomial is slid up from below until it "just touches"
some channel of the spectrum. Call this channel "i". The maximum
height of the polynomial is thus


                            (e(i) - e(j))**n
    height(j) = max ( b(j) +  --------------  )
                 i               width**n


where width is a user_specified parameter.

3) Once the value of height(i) is known the polynomial is fitted. The
background counts in each channel are then determined from:


                               (e(i) - e(j))**n
    bgr(j) = max ( height(i) + --------------  )
              i                     width**n


bgr(j) is thus the maximum counts for any of the concave down
polynomials passing though channel j.

Before the concave-down polynomials are fitted the spectrum at each
channel it is possible to subtract out a straight line which is
tangent to the spectrum at that channel. Use the TANGENT qualifier to
do this. This is equivalent to fitting a "tilted" polynomial whose
apex is tangent to the spectrum at that channel. By fitting
polynomials which are tangent rather than vertical the background fit
is much improved on spectra with steep slopes.

Input Parameter Fields:

    width (double, variable):
        Specifies the width of the polynomials which are concave downward.
        The bottom_width is the full width in energy units at which the
        magnitude of the polynomial is 100 counts. The default is 4.

    exponent (int):
        Specifies the power of polynomial which is used. The power must be
        an integer. The default is 2, i.e. parabolas. Higher exponents,
        for example EXPONENT=4, results in polynomials with flatter tops
        and steeper sides, which can better fit spectra with steeply
        sloping backgrounds.

    compress (int):
        Compression factor to apply before fitting the background.
        Default=4, which means, for example, that a 2048 channel spectrum
        will be rebinned to 512 channels before fitting.
        The compression is done on a temporary copy of the input spectrum,
        so the input spectrum itself is unchanged.
        The algorithm works best if the spectrum is compressed before it
        is fitted. There are two reasons for this. First, the background
        is constrained to never be larger than the data itself. If the
        spectrum has negative noise spikes they will cause the fit to be
        too low. Compression will smooth out such noise spikes.
        Second, the algorithm requires about 3*N**2 operations, so the time
        required grows rapidly with the size of the input spectrum. On a
        200 MHz Pentium it takes about 3 seconds to fit a 2048 channel
        spectrum with COMPRESS=1 (no compression), but only 0.2 seconds
        with COMPRESS=4 (the default).

        Note - compress needs a data array that integer divisible.

    tangent (True/False):
        Specifies that the polynomials are to be tangent to the slope of the
        spectrum. The default is vertical polynomials. This option works
        best on steeply sloping spectra. It has trouble in spectra with
        big peaks because the polynomials are very tilted up inside the
        peaks. Default is False

Inputs to calc()
    data:
       The raw data to fit the background

    slope:
        Slope for the conversion from channel number to energy.
        i.e. the slope from calibration

"""


import numpy as np

REFERENCE_AMPL=100.
TINY = 1.E-20
HUGE = 1.E20
MAX_TANGENT=2

def compress_array(array, compress):
    """
    Compresses an 1-D array by the integer factor compress.
    near equivalent of IDL's 'rebin'....
    """

    if len(array) % compress != 0:
        ## Trims array to be divisible by compress factor
        rng_min = int( (len(array) % compress ) / 2)
        rng_max = int( len(array) / compress ) * compress + 1
        array = array[rng_min:rng_max]

    nsize = int(len(array)/compress)
    temp = np.resize(array, (nsize, compress))
    return np.sum(temp, 1)/compress


def expand_array(array, expand, sample=0):
    """
    Expands an 1-D array by the integer factor expand.

    if 'sample' is 1 the new array is created with sampling,
    if 0 then the new array is created via interpolation (default)
    Temporary fix until the equivalent of IDL's 'rebin' is found.
    """
    if expand == 1:
        return array
    if sample == 1:
        return np.repeat(array, expand)

    kernel = np.ones(expand)/expand
    # The following mimic the behavior of IDL's rebin when expanding
    temp = np.convolve(np.repeat(array, expand), kernel, mode=2)
    # Discard the first "expand-1" entries
    temp = temp[expand-1:]
    # Replace the last "expand" entries with the last entry of original
    for i in range(1,expand):
        temp[-i]=array[-1]
    return temp

class XrayBackground:
    '''
    Class defining a spectrum background

    Attributes:
    -----------
    These may be set by kw argument upon initialization.


    * width      = 4.0   # Width
    * exponent   = 2     # Exponent
    * compress   = 2     # Compress
    * tangent    = False # Tangent flag
    '''

    def __init__(self, data=None, width=4, slope=1.0, exponent=2, compress=2,
                 tangent=False, type_int=False, data_type='xrf'):

        if data_type == 'xrf': type_int = True

        self.bgr          = []
        self.width = width
        self.compress     = compress
        self.exponent     = exponent
        self.tangent      = tangent

        self.info = {'width': width, 'compress': compress,
                     'exponent': exponent, 'tangent': tangent}

        self.data = data
        if data is not None:
            self.calc(data, slope=slope, type_int=type_int)

    def calc(self, data=None, slope=1.0, type_int=False):
        '''compute background

        Parameters:
        -----------
        * data is the spectrum
        * slope is the slope of conversion channels to energy
        '''

        if data is None:
            data = self.data

        width    = self.width
        exponent = self.exponent
        tangent  = self.tangent
        compress = self.compress

        nchans   = len(data)
        self.bgr = np.zeros(nchans, dtype=np.int)
        scratch  = data[:]

        # Compress scratch spectrum
        if compress > 1:
            tmp = compress_array(scratch, compress)
            if tmp is None:
                compress = 1
            else:
                scratch = tmp
                slope = slope * compress
                nchans = len(scratch) #nchans / compress

        # Copy scratch spectrum to background spectrum
        bckgnd = scratch[:]

        # Find maximum counts in input spectrum. This information is used to
        # limit the size of the function lookup table
        max_counts = max(scratch)

        # Fit functions which come up from below
        bckgnd = np.arange(nchans, dtype=np.float) - HUGE

        denom = max(TINY, (width / (2. * slope)**exponent))

        indices     = np.arange(nchans*2+1, dtype=np.float) - nchans
        power_funct = indices**exponent  * (REFERENCE_AMPL / denom)
        power_funct = np.compress((power_funct <= max_counts), power_funct)
        max_index   = int(len(power_funct)/2 - 1)
        for chan in range(nchans-1):
            tan_slope = 0.
            if tangent:
                # Find slope of tangent to spectrum at this channel
                chan0  = max((chan - MAX_TANGENT), 0)
                chan1  = min((chan + MAX_TANGENT), (nchans-1))
                denom  = chan - np.arange(chan1 - chan0 + 1, dtype=np.float)
                # is this correct?
                denom   = max(max(denom), 1)
                tan_slope = (scratch[chan] - scratch[chan0:chan1+1]) / denom
                tan_slope = np.sum(tan_slope) / (chan1 - chan0)

            chan0 = int(max((chan - max_index), 0))
            chan1 = int(min((chan + max_index), (nchans-1)))
            chan1 = max(chan1, chan0)
            nc    = chan1 - chan0 + 1
            lin_offset = scratch[chan] + (np.arange(float(nc)) - nc/2) * tan_slope

            # Find the maximum height of a function centered on this channel
            # such that it is never higher than the counts in any channel
            f      = int(chan0 - chan + max_index)
            l      = int(chan1 - chan + max_index)
            test   = scratch[chan0:chan1+1] - lin_offset + power_funct[f:l+1]
            height = min(test)

            # We now have the function height. Set the background to the
            # height of the maximum function amplitude at each channel

            test = height + lin_offset - power_funct[f:l+1]
            sub  = bckgnd[chan0:chan1+1]
            bckgnd[chan0:chan1+1] = np.maximum(sub, test)

        # Expand spectrum
        if compress > 1:
            bckgnd = expand_array(bckgnd, compress)

        ## Set background to be of type integer
        if type_int:
            bckgnd = bckgnd.astype(int)

        ## No negative values in background
        bckgnd[np.where(bckgnd <= 0)] = 0

        self.bgr = bckgnd



#!/usr/bin/env python
"""
   feffit sums Feff paths to match xafs data
"""
from collections import Iterable
from copy import copy, deepcopy
from functools import partial
import numpy as np
from numpy import array, arange, interp, pi, zeros, sqrt, concatenate

from scipy.optimize import leastsq as scipy_leastsq

from lmfit import Parameters, Parameter, Minimizer, fit_report
from lmfit.printfuncs import gformat as gformat

from larch import Group, isNamedClass

from ..math import index_of, realimag, complex_phase
from ..fitting import (correlated_values, eval_stderr,
                       group2params, params2group, isParameter)

from .xafsutils import set_xafsGroup
from .xafsft import xftf_fast, xftr_fast, ftwindow
from .sigma2_models import sigma2_correldebye, sigma2_debye
from .feffdat import PATHPAR_FMT, FeffPathGroup, ff2chi

class TransformGroup(Group):
    """A Group of transform parameters.
    The apply() method will return the result of applying the transform,
    ready to use in a Fit.   This caches the FT windows (k and r windows)
    and assumes that once created (not None), these do not need to be
    recalculated....

    That is: don't simply change the parameters and expect different results.
    If you do change parameters, reset kwin / rwin to None to cause them to be
    recalculated.
    """
    def __init__(self, kmin=0, kmax=20, kweight=2, dk=4, dk2=None,
                 window='kaiser', nfft=2048, kstep=0.05,
                 rmin = 0, rmax=10, dr=0, dr2=None, rwindow='hanning',
                 fitspace='r', wavelet_mask=None, _larch=None, **kws):
        Group.__init__(self, **kws)
        self.kmin = kmin
        self.kmax = kmax
        self.kweight = kweight
        if 'kw' in kws:
            self.kweight = kws['kw']

        self.dk = dk
        self.dk2 = dk2
        self.window = window
        self.rmin = rmin
        self.rmax = rmax
        self.dr = dr
        self.dr2 = dr2
        if dr2 is None: self.dr2 = self.dr
        self.rwindow = rwindow
        self.__nfft = 0
        self.__kstep = None
        self.nfft  = nfft
        self.kstep = kstep
        self.rstep = pi/(self.kstep*self.nfft)

        self.fitspace = fitspace
        self.wavelet_mask = wavelet_mask
        self._cauchymask = None

        self._larch = _larch

        self.kwin = None
        self.rwin = None
        self.make_karrays()

    def __repr__(self):
        return '<FeffitTransform Group: %s>' % self.__name__

    def __copy__(self):
        return TransformGroup(kmin=self.kmin, kmax=self.kmax,
                              kweight=self.kweight, dk=self.dk, dk2=self.dk2,
                              window=self.window, kstep=self.kstep,
                              rmin=self.rmin, rmax=self.rmax,
                              dr=self.dr, dr2=self.dr2,
                              rwindow=self.rwindow, nfft=self.nfft,
                              fitspace=self.fitspace,
                              wavelet_mask=self.wavelet_mask,
                              _larch=self._larch)

    def __deepcopy__(self, memo):
        return TransformGroup(kmin=self.kmin, kmax=self.kmax,
                              kweight=self.kweight, dk=self.dk, dk2=self.dk2,
                              window=self.window, kstep=self.kstep,
                              rmin=self.rmin, rmax=self.rmax,
                              dr=self.dr, dr2=self.dr2,
                              rwindow=self.rwindow, nfft=self.nfft,
                              fitspace=self.fitspace,
                              wavelet_mask=self.wavelet_mask,
                              _larch=self._larch)

    def make_karrays(self, k=None, chi=None):
        "this should be run in kstep or nfft changes"
        if self.kstep == self.__kstep and self.nfft == self.__nfft:
            return
        self.__kstep = self.kstep
        self.__nfft = self.nfft

        self.rstep = pi/(self.kstep*self.nfft)
        self.k_ = self.kstep * arange(self.nfft, dtype='float64')
        self.r_ = self.rstep * arange(self.nfft, dtype='float64')

    def _xafsft(self, chi, group=None, rmax_out=10, **kws):
        "returns "
        for key, val in kws:
            if key == 'kw':
                key = 'kweight'
            setattr(self, key, val)
        self.make_karrays()

        out = self.fftf(chi)

        irmax = int(min(self.nfft/2, 1.01 + rmax_out/self.rstep))

        group = set_xafsGroup(group, _larch=self._larch)
        r   = self.rstep * arange(irmax)
        mag = sqrt(out.real**2 + out.imag**2)
        group.kwin  =  self.kwin[:len(chi)]
        group.r    =  r[:irmax]
        group.chir =  out[:irmax]
        group.chir_mag =  mag[:irmax]
        group.chir_pha =  complex_phase(out[:irmax])
        group.chir_re  =  out.real[:irmax]
        group.chir_im  =  out.imag[:irmax]

    def get_kweight(self):
        "if kweight is a list/tuple, use only the first one here"
        if isinstance(self.kweight, Iterable):
            return self.kweight[0]
        return self.kweight

    def fftf(self, chi, kweight=None):
        """ forward FT -- meant to be used internally.
        chi must be on self.k_ grid"""
        if self.kstep != self.__kstep or self.nfft != self.__nfft:
            self.make_karrays()
        if self.kwin is None:
            self.kwin = ftwindow(self.k_, xmin=self.kmin, xmax=self.kmax,
                                 dx=self.dk, dx2=self.dk2, window=self.window)
        if kweight is None:
            kweight = self.get_kweight()
        cx = chi * self.kwin[:len(chi)] * self.k_[:len(chi)]**kweight
        return xftf_fast(cx, kstep=self.kstep, nfft=self.nfft)

    def fftr(self, chir):
        " reverse FT -- meant to be used internally"
        if self.kstep != self.__kstep or self.nfft != self.__nfft:
            self.make_karrays()
        if self.rwin is None:
            self.rwin = ftwindow(self.r_, xmin=self.rmin, xmax=self.rmax,
                                 dx=self.dr, dx2=self.dr2, window=self.rwindow)

        cx = chir * self.rwin[:len(chir)]
        return xftr_fast(cx, kstep=self.kstep, nfft=self.nfft)


    def make_cwt_arrays(self, nkpts, nrpts):
        if self.kstep != self.__kstep or self.nfft != self.__nfft:
            self.make_karrays()
        if self.kwin is None:
            self.kwin = ftwindow(self.k_, xmin=self.kmin, xmax=self.kmax,
                                 dx=self.dk, dx2=self.dk2, window=self.window)

        if self._cauchymask is None:
            if self.wavelet_mask is not None:
                self._cauchymask = self.wavelet_mask
            else:
                ikmin = int(max(0, 0.01 + self.kmin/self.kstep))
                ikmax = int(min(self.nfft/2,  0.01 + self.kmax/self.kstep))
                irmin = int(max(0, 0.01 + self.rmin/self.rstep))
                irmax = int(min(self.nfft/2,  0.01 + self.rmax/self.rstep))
                cm = np.zeros(nrpts*nkpts, dtype='int').reshape(nrpts, nkpts)
                cm[irmin:irmax, ikmin:ikmax] = 1
                self._cauchymask = cm
                self._cauchyslice =(slice(irmin, irmax), slice(ikmin, ikmax))

    def cwt(self, chi, rmax=None, kweight=None):
        """cauchy wavelet transform -- meant to be used internally"""
        if self.kstep != self.__kstep or self.nfft != self.__nfft:
            self.make_karrays()
        nkpts = len(chi)
        nrpts = int(np.round(self.rmax/self.rstep))
        if self.kwin is None:
            self.make_cwt_arrays(nkpts, nrpts)

        omega = pi*np.arange(self.nfft)/(self.kstep*self.nfft)

        if kweight is None:
            kweight = self.get_kweight()
        if kweight != 0:
            chi = chi * self.kwin[:len(chi)] * self.k_[:len(chi)]**kweight

        if rmax is not None:
            self.rmax = rmax

        chix   = np.zeros(int(self.nfft/2)) * self.kstep
        chix[:nkpts] = chi
        chix   = chix[:int(self.nfft/2)]
        _ffchi = np.fft.fft(chix, n=2*self.nfft)[:self.nfft]

        nrpts = int(np.round(self.rmax/self.rstep))
        r   = self.rstep * arange(nrpts)
        r[0] = 1.e-19
        alpha = nrpts/(2*r)

        self.make_cwt_arrays(nkpts, nrpts)

        cauchy_sum = np.log(2*pi) - np.log(1.0+np.arange(nrpts)).sum()

        out = np.zeros(nrpts*nkpts, dtype='complex128').reshape(nrpts, nkpts)

        for i in range(nrpts):
            aom = alpha[i]*omega
            filt = cauchy_sum + nrpts*np.log(aom) - aom
            out[i, :] = np.fft.ifft(np.exp(filt)*_ffchi, 2*self.nfft)[:nkpts]

        return (out*self._cauchymask)[self._cauchyslice]

class FeffitDataSet(Group):
    def __init__(self, data=None, pathlist=None, transform=None,
                 epsilon_k=None, _larch=None, **kws):
        self._larch = _larch
        Group.__init__(self, **kws)

        self.pathlist = pathlist

        self.data = data
        if transform is None:
            transform = TransformGroup()
        self.transform = transform
        if epsilon_k is not None:
            self.data.epsilon_k = epsilon_k

        self.model = Group()
        self.model.k = None
        self.__chi = None
        self.__prepared = False

    def __repr__(self):
        return '<FeffitDataSet Group: %s>' % self.__name__

    def __copy__(self):
        return FeffitDataSet(data=copy(self.data),
                             pathlist=self.pathlist[:],
                             transform=copy(self.transform),
                             _larch=self._larch)

    def __deepcopy__(self, memo):
        return FeffitDataSet(data=deepcopy(self.data),
                             pathlist=self.pathlist[:],
                             transform=deepcopy(self.transform),
                             _larch=self._larch)

    def prepare_fit(self, params):
        """prepare for fit with this dataset"""
        trans = self.transform
        trans.make_karrays()
        ikmax = int(1.01 + max(self.data.k)/trans.kstep)

        # ikmax = index_of(trans.k_, max(self.data.k))
        self.model.k = trans.k_[:ikmax]
        self.__chi = interp(self.model.k, self.data.k, self.data.chi)
        self.n_idp = 1 + 2*(trans.rmax-trans.rmin)*(trans.kmax-trans.kmin)/pi

        if getattr(self.data, 'epsilon_k', None) is not None:
            eps_k = self.data.epsilon_k
            if isinstance(eps_k, np.ndarray):
                eps_k = interp(self.model.k, self.data.k, self.data.epsilon_k)
            self.set_epsilon_k(eps_k)
        else:
            self.estimate_noise(chi=self.__chi, rmin=15.0, rmax=30.0)
            # uncertainty in chi(k) from autobk or other source
            if hasattr(self.data, 'delta_chi'):
                if isinstance(self.epsilon_k, (list, tuple)):
                    eps_ave = 0.
                    for eps in self.epsilon_k:
                        eps_ave += eps
                    self.epsilon_k = eps_ave/len(self.epsilon_k)
                _dchi = interp(self.model.k, self.data.k, self.data.delta_chi)
                eps_k = np.sqrt(_dchi**2 + self.epsilon_k**2)
                self.set_epsilon_k(eps_k)

        # for each path in the pathlist, setup the Path Parameters
        # to use the current Parameters namespace
        for path in self.pathlist:
            path.create_path_params(params=params)
            if path.spline_coefs is None:
                path.create_spline_coefs()

        self.__prepared = True


    def estimate_noise(self, chi=None, rmin=15.0, rmax=30.0, all_kweights=True):
        """estimage noise in a chi spectrum from its high r components"""
        trans = self.transform
        trans.make_karrays()
        if chi is None: chi = self.__chi

        save = trans.rmin, trans.rmax, trans.fitspace

        all_kweights = all_kweights and isinstance(trans.kweight, Iterable)
        if all_kweights:
            chir = [trans.fftf(chi, kweight=kw) for kw in trans.kweight]
        else:
            chir = [trans.fftf(chi)]
        irmin = int(0.01 + rmin/trans.rstep)
        irmax = int(min(trans.nfft/2, 1.01 + rmax/trans.rstep))
        highr = [realimag(chir_[irmin:irmax]) for chir_ in chir]
        # get average of window function value, we will scale eps_r scale by this
        kwin_ave = trans.kwin.sum()*trans.kstep/(trans.kmax-trans.kmin)
        eps_r = [(sqrt((chi*chi).sum() / len(chi)) / kwin_ave) for chi in highr]
        eps_k = []
        # use Parseval's theorem to convert epsilon_r to epsilon_k,
        # compensating for kweight
        if all_kweights:
            kweights = trans.kweight[:]
        else:
            kweights = [trans.kweight]

        for i, kw in enumerate(kweights):
            w = 2 * kw + 1
            scale = sqrt((2*pi*w)/(trans.kstep*(trans.kmax**w - trans.kmin**w)))
            eps_k.append(scale*eps_r[i])

        trans.rmin, trans.rmax, trans.fitspace = save

        ## self.n_idp  = 1 + 2*(trans.rmax-trans.rmin)*(trans.kmax-trans.kmin)/pi
        self.epsilon_k = eps_k
        self.epsilon_r = eps_r
        if len(eps_r) == 1:
            self.epsilon_k = eps_k[0]
            self.epsilon_r = eps_r[0]
        if isinstance(eps_r, np.ndarray):
            self.epsilon_r = eps_r.mean()

    def set_epsilon_k(self, eps_k):
        """set epsilon_k and epsilon_r -- ucertainties in chi(k) and chi(R)"""
        trans = self.transform
        all_kweights = isinstance(trans.kweight, Iterable)
        if isinstance(trans.kweight, Iterable):
            self.epsilon_k = []
            self.epsilon_r = []
            for kw in trans.kweight:
                w = 2 * kw + 1
                scale = 2*sqrt((pi*w)/(trans.kstep*(trans.kmax**w - trans.kmin**w)))
                self.epsilon_k.append(eps_k)
                eps_r = eps_k / scale
                if isinstance(eps_r, np.ndarray): eps_r = eps_r.mean()
                self.epsilon_r.append(eps_r)

        else:
            w = 2 * trans.get_kweight() + 1
            scale = 2*sqrt((pi*w)/(trans.kstep*(trans.kmax**w - trans.kmin**w)))
            self.epsilon_k = eps_k
            eps_r = eps_k / scale
            if isinstance(eps_r, np.ndarray): eps_r = eps_r.mean()
            self.epsilon_r = eps_r


    def _residual(self, paramgroup, data_only=False, **kws):
        """return the residual for this data set
        residual = self.transform.apply(data_chi - model_chi)
        where model_chi is the result of ff2chi(pathlist)
        """
        if not isNamedClass(self.transform, TransformGroup):
            return
        if not self.__prepared:
            self.prepare_fit()

        ff2chi(self.pathlist, paramgroup=paramgroup, k=self.model.k,
                _larch=self._larch, group=self.model)

        eps_k = self.epsilon_k
        if isinstance(eps_k, np.ndarray):
            eps_k[np.where(eps_k<1.e-12)[0]] = 1.e-12

        diff  = (self.__chi - self.model.chi)
        if data_only:  # for extracting transformed data separately from residual
            diff  = self.__chi
        trans = self.transform
        k     = trans.k_[:len(diff)]

        all_kweights = isinstance(trans.kweight, Iterable)
        if trans.fitspace == 'k':
            iqmin = int(max(0, 0.01 + trans.kmin/trans.kstep))
            iqmax = int(min(trans.nfft/2,  0.01 + trans.kmax/trans.kstep))
            if all_kweights:
                out = []
                for i, kw in enumerate(trans.kweight):
                    out.append(((diff/eps_k[i])*k**kw)[iqmin:iqmax])
                return np.concatenate(out)
            else:
                return ((diff/eps_k) * k**trans.kweight)[iqmin:iqmax]
        elif trans.fitspace == 'w':
            if all_kweights:
                out = []
                for i, kw in enumerate(trans.kweight):
                    cwt = trans.cwt(diff/eps_k, kweight=kw)
                    out.append(realimag(cwt).ravel())
                return np.concatenate(out)
            else:
                cwt = trans.cwt(diff/eps_k, kweight=trans.kweight)
                return realimag(cwt).ravel()
        else: # 'r' space
            out = []
            if all_kweights:
                chir = [trans.fftf(diff, kweight=kw) for kw in trans.kweight]
                eps_r = self.epsilon_r
            else:
                chir = [trans.fftf(diff)]
                eps_r = [self.epsilon_r]
            if trans.fitspace == 'r':
                irmin = int(max(0, 0.01 + trans.rmin/trans.rstep))
                irmax = int(min(trans.nfft/2,  0.01 + trans.rmax/trans.rstep))
                for i, chir_ in enumerate(chir):
                    chir_ = chir_ / (eps_r[i])
                    out.append(realimag(chir_[irmin:irmax]))
            else:
                chiq = [trans.fftr(c)/eps for c, eps in zip(chir, eps_r)]
                iqmin = int(max(0, 0.01 + trans.kmin/trans.kstep))
                iqmax = int(min(trans.nfft/2,  0.01 + trans.kmax/trans.kstep))
                for chiq_ in chiq:
                    out.append( realimag(chiq_[iqmin:iqmax])[::2])
            return np.concatenate(out)

    def save_ffts(self, rmax_out=10, path_outputs=True):
        "save fft outputs"
        xft = self.transform._xafsft
        xft(self.__chi,   group=self.data,  rmax_out=rmax_out)
        xft(self.model.chi, group=self.model, rmax_out=rmax_out)
        if path_outputs:
            for p in self.pathlist:
                xft(p.chi, group=p, rmax_out=rmax_out)

def feffit_dataset(data=None, pathlist=None, transform=None,
                   epsilon_k=None, _larch=None):
    """create a Feffit Dataset group.

     Parameters:
     ------------
      data:      group containing experimental EXAFS (needs arrays 'k' and 'chi').
      pathlist:  list of FeffPath groups, as created from feffpath()
      transform: Feffit Transform group.
      epsilon_k: Uncertainty in data (either single value or array of
                 same length as data.k)

     Returns:
     ----------
      a Feffit Dataset group.


    """
    return FeffitDataSet(data=data, pathlist=pathlist,
                         transform=transform, _larch=_larch)

def feffit_transform(_larch=None, **kws):
    """create a feffit transform group

     Parameters:
     --------------
       fitspace: name of FT type for fit  ('r').
       kmin:     starting *k* for FT Window (0).
       kmax:     ending *k* for FT Window (20).
       dk:       tapering parameter for FT Window (4).
       dk2:      second tapering parameter for FT Window (None).
       window:   name of window type ('kaiser').
       nfft:     value to use for N_fft (2048).
       kstep:    value to use for delta_k (0.05).
       kweight:  exponent for weighting spectra by k^kweight (2).
       rmin:     starting *R* for Fit Range and/or reverse FT Window (0).
       rmax:     ending *R* for Fit Range and/or reverse FT Window (10).
       dr:       tapering parameter for reverse FT Window 0.
       rwindow:  name of window type for reverse FT Window ('kaiser').

     Returns:
     ----------
       a feffit transform group.

    """
    return TransformGroup(_larch=_larch, **kws)

def feffit(paramgroup, datasets, rmax_out=10, path_outputs=True, _larch=None, **kws):
    """execute a Feffit fit: a fit of feff paths to a list of datasets

    Parameters:
    ------------
      paramgroup:   group containing parameters for fit
      datasets:     Feffit Dataset group or list of Feffit Dataset group.
      rmax_out:     maximum R value to calculate output arrays.
      path_output:  Flag to set whether all Path outputs should be written.

    Returns:
    ---------
      a fit results group.  This will contain subgroups of:

        datasets: an array of FeffitDataSet groups used in the fit.
        params:   This will be identical to the input parameter group.
        fit:      an object which points to the low-level fit.

     Statistical parameters will be put into the params group.  Each
     dataset will have a 'data' and 'model' subgroup, each with arrays:
        k            wavenumber array of k
        chi          chi(k).
        kwin         window Omega(k) (length of input chi(k)).
        r            uniform array of R, out to rmax_out.
        chir         complex array of chi(R).
        chir_mag     magnitude of chi(R).
        chir_pha     phase of chi(R).
        chir_re      real part of chi(R).
        chir_im      imaginary part of chi(R).
    """


    def _resid(params, datasets=None, paramgroup=None, **kwargs):
        """ this is the residual function"""
        params2group(params, paramgroup)
        return concatenate([d._residual(paramgroup) for d in datasets])

    if isNamedClass(datasets, FeffitDataSet):
        datasets = [datasets]

    params = group2params(paramgroup)
    for ds in datasets:
        if not isNamedClass(ds, FeffitDataSet):
            print( "feffit needs a list of FeffitDataSets")
            return
        ds.prepare_fit(params=params)

    fit = Minimizer(_resid, params,
                    fcn_kws=dict(datasets=datasets,
                                 paramgroup=paramgroup),
                    scale_covar=True, **kws)

    result = fit.leastsq()

    params2group(result.params, paramgroup)
    dat = concatenate([d._residual(paramgroup, data_only=True) for d in datasets])

    n_idp = 0
    for ds in datasets:
        n_idp += ds.n_idp

    # here we rescale chi-square and reduced chi-square to n_idp
    npts =  len(result.residual)
    chi_square  = result.chisqr * n_idp*1.0 / npts
    chi_reduced = chi_square/(n_idp*1.0 - result.nvarys)
    rfactor = (result.residual**2).sum() / (dat**2).sum()
    # calculate 'aic', 'bic' rescaled to n_idp
    # note that neg2_loglikel is -2*log(likelihood)
    neg2_loglikel = n_idp * np.log(chi_square / n_idp)
    aic = neg2_loglikel + 2 * result.nvarys
    bic = neg2_loglikel + np.log(n_idp) * result.nvarys


    # With scale_covar = True, Minimizer() scales the uncertainties
    # by reduced chi-square assuming params.nfree is the correct value
    # for degrees-of-freedom. But n_idp-params.nvarys is a better measure,
    # so we rescale uncertainties here.

    covar = getattr(result, 'covar', None)
    # print("COVAR " , covar)
    if covar is not None:
        err_scale = (result.nfree / (n_idp - result.nvarys))
        for name in result.var_names:
            p = result.params[name]
            if isParameter(p) and p.vary:
                p.stderr *= sqrt(err_scale)

        # next, propagate uncertainties to constraints and path parameters.
        result.covar *= err_scale
        vsave, vbest = {}, []

        # 1. save current params
        for vname in result.var_names:
            par = result.params[vname]
            vsave[vname] = par
            vbest.append(par.value)

        # 2. get correlated uncertainties, set params accordingly
        uvars = correlated_values(vbest, result.covar)
        # 3. evaluate constrained params, save stderr
        for nam, obj in result.params.items():
            eval_stderr(obj, uvars,  result.var_names, result.params)

        # 3. evaluate path params, save stderr
        for ds in datasets:
            for p in ds.pathlist:
                p.store_feffdat()
                for pname in ('degen', 's02', 'e0', 'ei',
                              'deltar', 'sigma2', 'third', 'fourth'):
                    obj = p.params[PATHPAR_FMT % (pname, p.label)]
                    eval_stderr(obj, uvars,  result.var_names, result.params)

        # restore saved parameters again
        for vname in result.var_names:
            # setattr(params, vname, vsave[vname])
            params[vname] = vsave[vname]

        # clear any errors evaluting uncertainties
        if _larch is not None and (len(_larch.error) > 0):
            _larch.error = []

    # reset the parameters group with the newly updated uncertainties
    params2group(result.params, paramgroup)

    # here we create outputs arrays for chi(k), chi(r):
    for ds in datasets:
        ds.save_ffts(rmax_out=rmax_out, path_outputs=path_outputs)

    out = Group(name='feffit results', datasets=datasets,
                fitter=fit, fit_details=result, chi_square=chi_square,
                n_independent=n_idp, chi_reduced=chi_reduced,
                rfactor=rfactor, aic=aic, bic=bic, covar=covar)

    for attr in ('params', 'nvarys', 'nfree', 'ndata', 'var_names', 'nfev',
                 'success', 'errorbars', 'message', 'lmdif_message'):
        setattr(out, attr, getattr(result, attr, None))
    return out


def feffit_report(result, min_correl=0.1, with_paths=True, _larch=None):
    """return a printable report of fit for feffit

    Parameters:
    ------------
      result:      Feffit result, output group from feffit()
      min_correl:  minimum correlation to report [0.1]
      wit_paths:   boolean (True/False) for whether to list all paths [True]

    Returns:
    ---------
      printable string of report.

    """
    input_ok = False
    try:
        params = result.params
        datasets = result.datasets
        input_ok = True
    except:
        pass
    if not input_ok:
        print( 'must pass output of feffit()!')
        return

    path_labels = []
    for ds in datasets:
        path_labels.extend([p.label for p in ds.pathlist])
    topline = '=================== FEFFIT RESULTS ===================='
    header = '[[%s]]'
    varformat  = '   %12s = %s +/-%s   (init= %s)'
    fixformat  = '   %12s = %s (fixed)'
    exprformat = '   %12s = %s +/-%s  = \'%s\''
    out = [topline, header % 'Statistics']

    out.append('   nvarys, npts       =  %i, %i' % (result.nvarys,
                                                   result.ndata))
    out.append('   n_independent      =  %.3f'  % (result.n_independent))
    out.append('   chi_square         = %s'  % gformat(result.chi_square))
    out.append('   reduced chi_square = %s'  % gformat(result.chi_reduced))
    out.append('   r-factor           = %s'  % gformat(result.rfactor))
    out.append('   Akaike info crit   = %s'  % gformat(result.aic))
    out.append('   Bayesian info crit = %s'  % gformat(result.bic))
    out.append(' ')

    if len(datasets) == 1:
        out.append(header % 'Data')
    else:
        out.append(header % 'Datasets (%i)' % len(datasets))
    for i, ds in enumerate(datasets):
        tr = ds.transform
        if len(datasets) > 1:
            out.append(' dataset %i:' % (i+1))
        if isinstance(tr.kweight, Iterable):
            if isinstance(ds.epsilon_k[0], np.ndarray):
                msg = []
                for eps in ds.epsilon_k:
                    msg.append('Array(mean=%s, std=%s)' % (gformat(eps.mean()).strip(),
                                                           gformat(eps.std()).strip()))
                eps_k = ', '.join(msg)
            else:
                eps_k = ', '.join([gformat(eps).strip() for eps in ds.epsilon_k])
            eps_r = ', '.join([gformat(eps).strip() for eps in ds.epsilon_r])
            kweigh = ', '.join(['%i' % kwe for kwe in tr.kweight])
        else:
            if isinstance(ds.epsilon_k, np.ndarray):
                eps_k = 'Array(mean=%s, std=%s)' % (gformat(ds.epsilon_k.mean()).strip(),
                                                    gformat(ds.epsilon_k.std()).strip())
            else:
                eps_k = gformat(ds.epsilon_k)
            eps_r = gformat(ds.epsilon_r).strip()
            kweigh = '%i' % tr.kweight
        out.append('   fit space          = \'%s\''  % (tr.fitspace))
        out.append('   r-range            = %.3f, %.3f' % (tr.rmin, tr.rmax))
        out.append('   k-range            = %.3f, %.3f' % (tr.kmin, tr.kmax))
        kwin = '   k window, dk       = \'%s\', %.3f'   % (tr.window, tr.dk)
        if tr.dk2 is not None:
            kwin = "%s, %.3f" % (kwin, tr.dk2)
        out.append(kwin)
        pathfiles = [p.filename for p in ds.pathlist]
        out.append('   paths used in fit  = %s' % (repr(pathfiles)))
        out.append('   k-weight           = %s' % kweigh)
        out.append('   epsilon_k          = %s'  % eps_k)
        out.append('   epsilon_r          = %s'  % eps_r)
        out.append('   n_independent      = %.3f'  % (ds.n_idp))
        #
    out.append(' ')
    out.append(header % 'Variables')
    for name, par in params.items():
        if any([name.endswith('__%s' % p_lab) for p_lab in path_labels]):
            continue
        if len(name) < 14:
            name = (name + ' '*14)[:14]

        if isParameter(par):
            if par.vary:
                stderr = 'unknown'
                if par.stderr is not None:
                    stderr = gformat(par.stderr)
                out.append(varformat % (name, gformat(par.value),
                                        stderr, gformat(par.init_value)))

            elif par.expr is not None:
                stderr = 'unknown'
                if par.stderr is not None:
                    stderr = gformat(par.stderr)
                out.append(exprformat % (name, gformat(par.value),
                                         stderr, par.expr))
            else:
                out.append(fixformat % (name, gformat(par.value)))

    covar_vars = result.var_names
    if len(covar_vars) > 0:
        out.append(' ')
        out.append(header % 'Correlations' +
                   '    (unreported correlations are < % .3f)' % min_correl)
        correls = {}
        for i, name in enumerate(covar_vars):
            par = params[name]
            if not par.vary:
                continue
            if hasattr(par, 'correl') and par.correl is not None:
                for name2 in covar_vars[i+1:]:
                    if name != name2 and name2 in par.correl:
                        correls["%s, %s" % (name, name2)] = par.correl[name2]

        sort_correl = sorted(correls.items(), key=lambda it: abs(it[1]))
        sort_correl.reverse()
        for name, val in sort_correl:
            if abs(val) < min_correl:
                break
            if len(name) < 20:
                name = (name + ' '*20)[:20]
            out.append('   %s = % .3f' % (name, val))

    if with_paths:
        out.append(' ')
        out.append(header % 'Paths')
        for ids, ds in enumerate(datasets):
            if len(datasets) > 1:
                out.append(' dataset %i:' % (ids+1))
            for p in ds.pathlist:
                out.append('%s\n' % p.report())
    out.append('='*len(topline))
    return '\n'.join(out)



'''
This module defines a device-independent XRD class.

Authors/Modifications:
----------------------
* Margaret Koker, koker@cars.uchicago.edu
* modeled after MCA class
'''

##########################################################################
# IMPORT PYTHON PACKAGES

import os
import numpy as np
import larch

from .xrd_tools import (d_from_q,d_from_twth,twth_from_d,twth_from_q,
                        q_from_d,q_from_twth,E_from_lambda,lambda_from_E)
from .xrd_pyFAI import integrate_xrd,calc_cake
from .xrd_bgr import xrd_background
from .xrd_fitting import peakfinder,peaklocater,peakfilter,peakfitter
from larch.io import tifffile

##########################################################################
# CLASSES

class xrd1d(larch.Group):
    '''
    1D XRD data class

    --> all length units in m (unless otherwise noted)

    Attributes:
    ------------
    * self.filename      = 'CeO2_Allende.xy'      # file containing x-y data
    * self.label         = 'Data: CeO2_Allende'   # name of data set

    # Data parameters
    * self.I             = None or array          # intensity (units: counts)
    * self.q             = None or array          # inverse lattice spacing (units: 1/A)
    * self.d             = None or array          # lattice spacing (units: A)
    * self.twth          = None or array          # 2-theta (units: degrees)

    # Calibration/experiemental parameters
    * self.wavelength    = 0.688801071778         # incident x-ray wavelength (units: A)
    * self.energy        = 18.000                 # incident x-ray energy (units: keV)
    * self.distance      = 0.49212785             # distance from sample to detector
    * self.poni          = [0.2286, 0.2547]       # center/point of normal incidence
    * self.rotation      = [0.010, -0.00857, 0.0] # detector rotations (units: radians)
    * self.pixelsize     = [0.0004, 0.0004]       # size of pixels
    * self.splinefile    = None                   # spline file for detector
    * self.polarization  = None                   # polarization of detector
    * self.normalization = 1.0                    # normalization factor for detector

    # Data fitting parameters
    * self.uvw           = [0.313, -0.109, 0.019] # instrumental broadening parameters
    * self.D             = None                   # particle size broadening (units: A)
    * self.pki           = [8, 254, 3664]         # list of peak indecides

    * self.imin          = None                   # range for trimmed data
    * self.imax          = None                   # range for trimmed data
    * self.bkgd          = None                   # fit background for data

    * self.matches       = None                   # list of amcsd matches from database

    * self.xrd2d         = None                   # 2D data
    * self.cake          = None                   # 2D cake

    mkak 2017.03.15
    '''

    def __init__(self, file=None, label=None, x=None, xtype=None, I=None,
                 wavelength=None, energy=None):

        self.filename = file
        self.label    = label

        self.energy     = energy
        self.wavelength = wavelength

        if energy is None and wavelength is None:
            self.energy = 19.0
            self.wavelength = lambda_from_E(self.energy)
        else:
            if self.energy is None:     self.energy = E_from_lambda(self.wavelength)
            if self.wavelength is None: self.wavelength = lambda_from_E(self.energy)

        if file is not None:
            self.xrd_from_file(file)
        else:

            ## Default values
            self.distance      = None
            self.poni          = None
            self.rotation      = None
            self.pixelsize     = None
            self.splinefile    = None
            self.polarization  = None
            self.normalization = None

            if I is not None and x is not None:
                self.xrd_from_2d([x,I],xtype)
                self.bkgd = np.zeros(np.shape(self.I))
            else:
                self.q    = None
                self.twth = None
                self.d    = None
                self.I    = None
                self.bkgd = None


        ## Analysis parameters - set defaults
        self.uvw = None
        self.D   = None
        self.pki = []

        self.imin = None
        self.imax = None

        self.matches = None

        self.xrd2d   = None
        self.cake    = None

        larch.Group.__init__(self)


    def xrd_from_2d(self,xy,xtype,verbose=True):
        self.set_xy_data(xy, xtype)

    def xrd_from_file(self, filename, verbose=True):

        try:
            from ..xrmmap import read1DXRDFile
            head,dat = read1DXRDFile(filename)
            if verbose:
                print('Opening xrd data file: %s' % os.path.split(filename)[-1])
            if len(head) < 4:
                print('WARNING: Using default energy for data. None given in file.')
        except:
           print('incorrect xy file format: %s' % os.path.split(filename)[-1])
           return

        if self.label is None: self.label = os.path.split(filename)[-1]

        ## header info
        for line in head:
            import re
            line = re.sub(',','',line)

            if 'SplineFile' in line:
                self.splinefile = line.split()[-1]
            if 'PixelSize' in line:
                self.pixelsize = [float(line.split()[-3]),float(line.split()[-2])]
            if 'PONI' in line:
                self.poni = [float(line.split()[2]),float(line.split()[3])]
            if 'Distance Sample to Detector' in line:
                self.distance = float(line.split()[-2])
            if 'Rotations' in line:
                self.rotation = [float(line.split()[2]),float(line.split()[3]),float(line.split()[4])]

            if 'Wavelength' in line:
                try:
                    self.wavelength = float(line.split()[-1])*1e10
                except:
                    self.wavelength = float(line.split()[-2])*1e10
                self.energy = E_from_lambda(self.wavelength)
            if 'Polarization' in line:
                if line.split()[-1] != 'None': self.polarization = float(line.split()[-1])
            if 'Normalization' in line:
                try:
                    value = float(line.split()[-1])
                except:
                    value = 1.0
                self.normalization = value

            if 'q_' in line or '2th_' in line:
                xtype = line.split()[1]

        ## data
        self.set_xy_data(dat,xtype)

    def set_xy_data(self, xy, xtype):
        if xy is not None:
            xy = np.array(xy)
            if xy.shape[0] > xy.shape[1]:
                x,y = np.split(xy,2,axis=1)
            else:
                x,y = np.split(xy, 2, axis=0)
            self.q, self.twth, self.d = calculate_xvalues(x, xtype, self.wavelength)
            self.I = np.array(y).squeeze()

            self.imin,self.imax = 0,len(self.q)
            self.bkgd = np.zeros(np.shape(self.I))

    def plot(self,reset=False,bkgd=False):

        if reset: self.imin,self.imax = 0,len(self.I)
        if bkgd:
            return [self.q[self.imin:self.imax],
                    self.twth[self.imin:self.imax],
                    self.d[self.imin:self.imax],
                    self.I[self.imin:self.imax]-self.bkgd,
                    self.bkgd]
        return [self.q[self.imin:self.imax],
                self.twth[self.imin:self.imax],
                self.d[self.imin:self.imax],
                self.I[self.imin:self.imax],
                self.bkgd]

    def reset_bkgd(self):
         self.bkgd = np.zeros(np.shape(self.I))

    def slct_xaxis(self,xtype='',xi=None):

        if xtype.startswith('q') or xi == 0:
            x = self.q
        elif xtype.startswith('2th') or xi == 1:
            x = self.twth
        elif xtype.startswith('d') or xi == 2:
            x = self.d
        else:
            print('The provided x-axis label (%s or &i) not correct.' % (xtype,xi))
            return

        return x

    def set_trim(self,xmin,xmax,xtype='',xi=None):

        x = self.slct_xaxis(xtype=xtype,xi=xi)

        self.imin,self.imax = 0,len(x)-1
        if xmin > np.min(x):
            self.imin = (np.abs(x-xmin)).argmin()
        if xmax < np.max(x):
            self.imax = (np.abs(x-xmax)).argmin()

    def all_data(self,reset=False,bkgd=False):

        if reset: self.imin,self.imax = 0,len(self.I)
        if len(self.I[self.imin:self.imax]) != len(self.bkgd):
            self.bkgd = np.zeros(len(self.I[self.imin:self.imax]))
        if bkgd:
            return [self.q[self.imin:self.imax],
                    self.twth[self.imin:self.imax],
                    self.d[self.imin:self.imax],
                    self.I[self.imin:self.imax]-self.bkgd,
                    self.bkgd]
        return [self.q[self.imin:self.imax],
                self.twth[self.imin:self.imax],
                self.d[self.imin:self.imax],
                self.I[self.imin:self.imax],
                self.bkgd]

    def fit_background(self,**kwargs):

        x,y = self.q[self.imin:self.imax],self.I[self.imin:self.imax]
        self.bkgd = xrd_background(x,y,**kwargs)

        while len(self.bkgd) < len(y):
            self.bkgd = np.append(self.bkgd,self.bkgd[-1])

    def find_peaks(self,bkgd=False,threshold=None,**kwargs):

        all_data = np.array(self.all_data(bkgd=bkgd))


        self.pki = peakfinder(all_data[3],**kwargs)
        if threshold is not None:
            self.pki = peakfilter(threshold,self.pki,all_data[3])

        pk_data = np.zeros((5,len(self.pki)))
        for i,pki in enumerate(self.pki): pk_data[:,i] = all_data[:,pki]

#     def refine_peaks(self,trim=False,bkgd=False):
#
#         q,twth,d,I = self.trim_all(trim,bkgd)
#
#         pktwth,pkfwhm,self.Ipks = peakfitter(self.pki,twth,I,fittype='double')
#         #self.peaks = zip(pkfwhm,pkI)
#
#         self.qpks,self.twthpks,self.dpks = calculate_xvalues(pktwth,'2th',self.wavelength)

#     def fit_pattern(self):
#         ## This isn't written yet
#         ## mkak 2017.04.03
#         fit = np.zeros(len(self.I))
#         for i,j in enumerate(self.pki):
#             a = None



def read_xrd_data(filepath):

    if not os.path.exists(filepath):
        return

    try:
        data = np.array(tifffile.imread(filepath))
    except: # TypeError:
        try:
            from ..xrmmap import read_xrd_netcdf
            data = np.array(read_xrd_netcdf(filepath))
        except:
            try:
                data = xrd1d(file=filepath).I
            except:
                return
    return data


class XRD(larch.Group):
    '''
    X-Ray Diffraction (XRD) class

    Attributes:
    -----------
    * self.name        = 'xrd'  # Name of the object
    * self.xpix        = 2048   # Number of x pixels
    * self.ypix        = 2048   # Number of y pixels
    * self.data2D      = None   # 2D XRD data
    * self.data1D      = None   # 1D XRD data

    Notes:
    ------

    mkak 2016.08.20
    '''

    def __init__(self, data2D=None, xpixels=2048, ypixels=2048,
                       data1D=None, nwedge=0, title=None,
                       steps=5001, name='xrd', filename=None,
                       calfile=None, energy=None, wavelength=None,
                       npixels=None, _larch=None, **kws):

        self.name    = name
        self.xpix    = xpixels
        self.ypix    = ypixels
        self.data2D  = data2D
        self.nwedge  = nwedge
        self.steps   = steps
        self.data1D  = data1D
        self.data2D  = data2D
        self.cake    = None

        self.calfile    = calfile

        if energy is None and wavelength is not None:
            self.wavelegth = wavelength
            self.energy = E_from_lambda(wavelength)
        elif energy is not None and wavelength is None:
            self.energy = energy
            self.wavelength = lambda_from_E(self.energy)
        else:
            self.energy     = energy
            self.wavelength = wavelength

        self.filename = filename
        self.title    = title
        self.npixels  = npixels

        larch.Group.__init__(self)

    def __repr__(self):
        if self.data2D is not None:
            form = "<2DXRD %s: pixels = %d, %d>"
            return form % (self.name, self.xpix, self.ypix)
        elif self.data1D is not None:
            form = "<1DXRD %s: channels = %d>"
            return form % (self.name, self.steps)
        else:
            form = "<no 1D or 2D XRD pattern given>"
            return form

    def add_environ(self, desc='', val='', addr=''):
        '''add an Environment setting'''
        if len(desc) > 0 and len(val) > 0:
            self.environ.append(Environment(desc=desc, val=val, addr=addr))


    def calc_1D(self,save=False,calccake=True,calc1d=True,verbose=False):

        kwargs = {'steps':self.steps}

        if save:
            file = self.save_1D()
            kwargs.update({'file':file})

        if os.path.exists(self.calfile):
            if len(self.data2D) > 0:
                self.data1D = integrate_xrd(self.data2D,self.calfile,
                                            verbose=verbose,**kwargs)
                self.cake = calc_cake(self.data2D,self.calfile, unit='q')
            else:
                if verbose:
                    print('No 2D XRD data provided.')
        else:
            if verbose:
                print('Could not locate file %s' % self.calfile)

    def save_1D(self,file=None):

        if file is None:
            counter = 1
            while os.path.exists('%s/%s_%03d.xy' % (os.getcwd(),self.title,counter)):
                counter += 1
            file = '%s/%s_%03d.xy' % (os.getcwd(),self.title,counter)

        return file


    def save_2D(self,file=None,verbose=False):

        if file is None:
            counter = 1
            while os.path.exists('%s/%s_%03d.tiff' % (os.getcwd(),self.title,counter)):
                counter += 1
            file = '%s/%s_%03d.tiff' % (os.getcwd(),self.title,counter)

        tifffile.imsave(file,self.data2D)



##########################################################################
# FUNCTIONS

def calculate_xvalues(x,xtype,wavelength):
    '''
    projects given x-axis onto q-, 2theta-, and d-axes

    x            :   list or array (expected units: 1/A, deg, or A)
    xtype        :   options 'q', '2th', or 'd'
    wavelength   :   incident x-ray wavelength (units: A)

    q, twth, d   :   returned with same dimensions as x (units: 1/A, deg, A)
    '''

    x = np.array(x).squeeze()
    if xtype.startswith('q'):

        q = x
        d = d_from_q(q)
        if wavelength is not None:
            twth = twth_from_q(q,wavelength)
        else:
            twth = np.zeros(len(q))

    elif xtype.startswith('2th'):

        twth = x
        if wavelength is not None:
            q = q_from_twth(twth,wavelength)
            d = d_from_twth(twth,wavelength)
        else:
            q = np.zeros(len(twth))
            d = np.zeros(len(twth))

    elif xtype.startswith('d'):

        d = x
        q = q_from_d(d)
        if wavelength is not None:
            twth = twth_from_d(d,wavelength)
        else:
            twth = np.zeros(len(d))

    else:
        print('The provided x-axis label (%s) not correct. Check data.' % xtype)
        return None,None,None


    return q,twth,d


def create_xrd(data2D=None, xpixels=2048, ypixels=2048,
               data1D=None, nwedge=0, steps=5001,
               name='xrd', _larch=None, **kws):

    '''
    create an XRD object

     Parameters:
     ------------
      data2D:   2D diffraction patterns
      data1D:   1D diffraction patterns
      xpixels:  number of x pixels
      ypixels:  number of y pixels

     Returns:
     ----------
      an XRD object

    '''
    return XRD(data2D=data2D, data1D=data1D, xpixels=xpixels, ypixels=ypixels,
               name=name, **kws)


def create_xrd1d(file, _larch=None, **kws):

    '''
    create an XRD object

     Parameters:
     ------------
      data2D:   2D diffraction patterns
      data1D:   1D diffraction patterns
      xpixels:  number of x pixels
      ypixels:  number of y pixels

     Returns:
     ----------
      an XRD object

    '''
    return xrd1d(file=file, **kws)
